from typing import Any, ClassVar, Dict, Generic, Iterator, List, Optional, Set, Union, cast

from vellum import ChatMessage, PromptBlock, PromptOutput
from vellum.client.types.prompt_parameters import PromptParameters
from vellum.client.types.prompt_settings import PromptSettings
from vellum.client.types.string_chat_message_content import StringChatMessageContent
from vellum.prompts.constants import DEFAULT_PROMPT_PARAMETERS
from vellum.workflows.context import execution_context, get_parent_context
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.events.node import NodeExecutionStreamingEvent
from vellum.workflows.events.workflow import is_workflow_event
from vellum.workflows.exceptions import NodeException
from vellum.workflows.expressions.coalesce_expression import CoalesceExpression
from vellum.workflows.graph.graph import Graph
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.nodes.displayable.tool_calling_node.state import ToolCallingState
from vellum.workflows.nodes.displayable.tool_calling_node.utils import (
    create_else_node,
    create_function_node,
    create_router_node,
    create_tool_prompt_node,
    get_function_name,
)
from vellum.workflows.outputs.base import BaseOutput, BaseOutputs
from vellum.workflows.references.output import OutputReference
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.types.core import EntityInputsInterface
from vellum.workflows.types.definition import MCPServer, MCPToolDefinition, Tool, ToolBase
from vellum.workflows.types.generics import StateType
from vellum.workflows.utils.functions import compile_mcp_tool_definition
from vellum.workflows.workflows.event_filters import all_workflow_event_filter


def _contains_reference_to_output(reference: BaseDescriptor, target_reference: OutputReference) -> bool:
    if reference == target_reference:
        return True
    if isinstance(reference, CoalesceExpression):
        return _contains_reference_to_output(reference.lhs, target_reference) or _contains_reference_to_output(
            reference.rhs, target_reference
        )
    return False


class ToolCallingNode(BaseNode[StateType], Generic[StateType]):
    """
    A Node that dynamically invokes the provided functions to the underlying Prompt

    Attributes:
        ml_model: str - The model to use for tool calling (e.g., "gpt-4o-mini")
        blocks: List[PromptBlock] - The prompt blocks to use (same format as InlinePromptNode)
        functions: List[Tool] - The functions that can be called
        prompt_inputs: Optional[EntityInputsInterface] - Mapping of input variable names to values
        parameters: PromptParameters - The parameters for the Prompt
        max_prompt_iterations: Optional[int] - Maximum number of prompt iterations before stopping
    """

    class Display(BaseNode.Display):
        icon = "vellum:icon:wrench"
        color = "gold"

    ml_model: ClassVar[str] = "gpt-4o-mini"
    blocks: ClassVar[List[Union[PromptBlock, Dict[str, Any]]]] = []
    functions: ClassVar[List[Tool]] = []
    prompt_inputs: ClassVar[Optional[EntityInputsInterface]] = None
    parameters: PromptParameters = DEFAULT_PROMPT_PARAMETERS
    max_prompt_iterations: ClassVar[Optional[int]] = 25
    settings: ClassVar[Optional[Union[PromptSettings, Dict[str, Any]]]] = None

    class Outputs(BaseOutputs):
        """
        The outputs of the ToolCallingNode.

        text: The final text response after tool calling
        chat_history: The complete chat history including tool calls
        """

        text: str
        chat_history: List[ChatMessage]

    def run(self) -> Iterator[BaseOutput]:
        """
        Run the tool calling workflow with streaming support.

        This dynamically builds a graph with router and function nodes,
        then executes the workflow and streams chat_history updates.
        """

        self._build_graph()

        with execution_context(parent_context=get_parent_context()):

            from vellum.workflows.workflows.base import BaseWorkflow

            class AgentWorkflow(BaseWorkflow[BaseInputs, ToolCallingState]):
                graph = self._graph
                is_dynamic = True

                class Outputs(BaseWorkflow.Outputs):
                    text: str = self.tool_prompt_node.Outputs.text
                    chat_history: List[ChatMessage] = ToolCallingState.chat_history
                    results: List[PromptOutput] = self.tool_prompt_node.Outputs.results

            subworkflow = AgentWorkflow(
                parent_state=self.state,
                context=WorkflowContext.create_from(self._context),
            )

            subworkflow_stream = subworkflow.stream(
                event_filter=all_workflow_event_filter,
                node_output_mocks=self._context._get_all_node_output_mocks(),
                event_max_size=self._context.event_max_size,
            )

        outputs: Optional[BaseOutputs] = None
        exception: Optional[NodeException] = None
        fulfilled_output_names: Set[str] = set()

        for event in subworkflow_stream:
            self._context._emit_subworkflow_event(event)

            if not is_workflow_event(event):
                continue
            if event.workflow_definition != AgentWorkflow:
                continue

            if event.name == "workflow.execution.streaming":
                if event.output.name == "results":
                    if event.output.is_streaming:
                        if isinstance(event.output.delta, str):
                            yield BaseOutput(name="text", delta=event.output.delta)

                            text_message = ChatMessage(
                                text=event.output.delta,
                                role="ASSISTANT",
                                content=StringChatMessageContent(type="STRING", value=event.output.delta),
                                source=None,
                            )
                            yield BaseOutput(
                                name="chat_history",
                                delta=[text_message],
                            )
                elif event.output.name == "chat_history":
                    if event.output.is_fulfilled:
                        fulfilled_output_names.add(event.output.name)
                    yield event.output
                elif event.output.name == "text":
                    if event.output.is_fulfilled:
                        fulfilled_output_names.add(event.output.name)
                    yield event.output
            elif event.name == "workflow.execution.fulfilled":
                outputs = event.outputs
            elif event.name == "workflow.execution.rejected":
                exception = NodeException.of(event.error)

        if exception:
            raise exception

        if outputs is None:
            raise NodeException(
                message="Expected to receive outputs from Tool Calling Workflow",
                code=WorkflowErrorCode.INVALID_OUTPUTS,
            )

        for output_descriptor, output_value in outputs:
            if output_descriptor.name not in fulfilled_output_names:
                yield BaseOutput(
                    name=output_descriptor.name,
                    value=output_value,
                )

    def _build_graph(self) -> None:
        # Get the process_parameters method if it exists on this class
        process_parameters_method = getattr(self.__class__, "process_parameters", None)
        process_blocks_method = getattr(self.__class__, "process_blocks", None)

        # Hydrate MCP servers upfront and replace them with tool definitions
        hydrated_functions: List[Union[ToolBase, MCPToolDefinition]] = []
        for function in self.functions:
            if isinstance(function, MCPServer):
                hydrated_functions.extend(compile_mcp_tool_definition(function))
            else:
                # After checking, function is ToolBase (not MCPServer)
                # Mypy doesn't narrow Union[ToolBase, MCPServer] to ToolBase, so we cast
                hydrated_functions.append(cast(ToolBase, function))

        self.tool_prompt_node = create_tool_prompt_node(
            ml_model=self.ml_model,
            blocks=self.blocks,
            functions=hydrated_functions,
            prompt_inputs=self.prompt_inputs,
            parameters=self.parameters,
            max_prompt_iterations=self.max_prompt_iterations,
            process_parameters_method=process_parameters_method,
            process_blocks_method=process_blocks_method,
            settings=self.settings,
        )

        # Create the router node (handles routing logic only)
        self.router_node = create_router_node(
            functions=hydrated_functions,
            tool_prompt_node=self.tool_prompt_node,
        )

        self._function_nodes = {}
        for hydrated_function in hydrated_functions:
            function_name = get_function_name(hydrated_function)
            self._function_nodes[function_name] = create_function_node(
                function=hydrated_function,
                tool_prompt_node=self.tool_prompt_node,
            )

        graph: Graph = self.tool_prompt_node >> self.router_node

        for function_name, FunctionNodeClass in self._function_nodes.items():
            router_port = getattr(self.router_node.Ports, function_name)
            function_subgraph = router_port >> FunctionNodeClass >> self.router_node
            graph._extend_edges(function_subgraph.edges)

        else_node = create_else_node(self.tool_prompt_node)
        default_port_graph = self.router_node.Ports.default >> {
            else_node.Ports.loop_to_router >> self.router_node,  # More outputs to process
            else_node.Ports.loop_to_prompt >> self.tool_prompt_node,  # Need new prompt iteration
            else_node.Ports.end,  # Finished
        }
        graph._extend_edges(default_port_graph.edges)

        self._graph = graph

    def __directly_emit_workflow_output__(
        self,
        event: NodeExecutionStreamingEvent,
        workflow_output_descriptor: OutputReference,
    ) -> bool:
        """
        Check if this ToolCallingNode should directly emit workflow output events.
        Similar to BasePromptNode, this allows streaming text and chat_history outputs through FinalOutputNode.
        """
        # Only handle text and chat_history outputs
        if event.output.name != "text":
            return False

        if not isinstance(event.output.delta, str) and not event.output.is_initiated:
            return False

        target_nodes = [e.to_node for port in self.Ports for e in port.edges if e.to_node.__simulates_workflow_output__]
        target_node_output = next(
            (
                o
                for target_node in target_nodes
                for o in target_node.Outputs
                if o == workflow_output_descriptor.instance
            ),
            None,
        )
        if not target_node_output:
            return False

        if not isinstance(target_node_output.instance, BaseDescriptor):
            return False

        return _contains_reference_to_output(target_node_output.instance, event.node_definition.Outputs.text)



from typing import Any, Union, get_args, get_origin


class TypeValidationError(ValueError):
    """Exception raised when type validation fails."""

    def __init__(self, declared_type: type, target_type: type) -> None:
        declared_type_name = getattr(declared_type, "__name__", str(declared_type))
        target_type_name = getattr(target_type, "__name__", str(target_type))
        super().__init__(
            f"Output type mismatch: declared type '{declared_type_name}' "
            f"but the 'value' Output has type(s) '{target_type_name}'. "
        )
        self.declared_type = declared_type
        self.target_type = target_type


def validate_target_type(declared_type: type, target_type: type) -> None:
    # If either type is Any, validation always passes
    if declared_type is Any or target_type is Any:
        return

    declared_origin = get_origin(declared_type)
    target_origin = get_origin(target_type)

    # Special case: both are Unions
    # Each target union member must be compatible with at least one declared union member
    if declared_origin is Union and target_origin is Union:
        declared_union_args = get_args(declared_type)
        target_union_args = get_args(target_type)
        for target_member in target_union_args:
            is_compatible = False
            for declared_member in declared_union_args:
                try:
                    validate_target_type(declared_member, target_member)
                    is_compatible = True
                    break
                except TypeValidationError:
                    continue
            if not is_compatible:
                # This target member is not compatible with any declared member
                raise TypeValidationError(declared_type, target_type)
        # All target members are compatible with at least one declared member
        return

    # If declared_type is a Union, target_type must be compatible with at least one union member
    if declared_origin is Union:
        declared_union_args = get_args(declared_type)
        for union_member in declared_union_args:
            try:
                validate_target_type(union_member, target_type)
                # If any union member is compatible, validation passes
                return
            except TypeValidationError:
                # This union member is incompatible, try the next one
                continue
        # None of the union members are compatible with target_type
        raise TypeValidationError(declared_type, target_type)

    # If target_type is a Union, ALL union members must be compatible with declared_type
    if target_origin is Union:
        target_union_args = get_args(target_type)
        for union_member in target_union_args:
            try:
                validate_target_type(declared_type, union_member)
            except TypeValidationError:
                # At least one union member is incompatible with declared_type
                raise TypeValidationError(declared_type, target_type)
        # All union members are compatible with declared_type
        return

    # Check for exact type match
    if target_type == declared_type:
        return

    # Check for subclass relationships
    try:
        if issubclass(target_type, declared_type) or issubclass(declared_type, target_type):
            return
    except TypeError:
        # Handle cases where types aren't classes (e.g., generic types)
        if str(target_type) == str(declared_type):
            return

    # Check for generic type compatibility
    declared_origin = get_origin(declared_type)

    if target_origin is None and declared_origin is None:
        # Neither is a generic type, and they don't match
        raise TypeValidationError(declared_type, target_type)

    if target_origin == declared_type or declared_origin == target_type or target_origin == declared_origin:
        target_args = get_args(target_type)
        declared_args = get_args(declared_type)

        if len(declared_args) == 0:
            return

        if len(target_args) != len(declared_args):
            raise TypeValidationError(declared_type, target_type)

        for target_arg, declared_arg in zip(target_args, declared_args):
            validate_target_type(declared_arg, target_arg)

        return

    # If we reach here, types don't match
    raise TypeValidationError(declared_type, target_type)



from vellum_ee.workflows.display.workflows.base_workflow_display import BaseWorkflowDisplay


def test_serialize_module__invalid_pdf_data_url():
    """
    Tests that serialization returns an error when sandbox.py has an invalid PDF data URL
    without wrecking the whole serialize process, and that valid dataset rows are still returned.
    """

    # GIVEN a workflow module with a sandbox.py that has one invalid and one valid PDF data URL
    module = "tests.workflows.test_sandbox_invalid_pdf_data_url"

    # WHEN we serialize the module
    result = BaseWorkflowDisplay.serialize_module(module)

    # THEN the result should still contain a valid exec_config
    assert result.exec_config is not None
    assert isinstance(result.exec_config, dict)

    # AND the result should contain exactly one error about the invalid PDF data URL
    assert len(result.errors) == 1
    error_message = result.errors[0].message

    # AND the error message should include the dataset row label
    assert 'Dataset row "Scenario with invalid PDF"' in error_message

    # AND the error message should include the input name
    assert 'input "document"' in error_message

    # AND the error message should mention the invalid base64 encoding
    assert "Invalid base64 encoding in PDF data URL" in error_message

    # AND the result should contain both dataset rows
    assert result.dataset is not None
    assert len(result.dataset) == 2

    # AND the first row should have the valid input (name) serialized but the invalid input (document) skipped
    invalid_row = result.dataset[0]
    assert invalid_row["label"] == "Scenario with invalid PDF"
    assert "document" not in invalid_row["inputs"]
    assert "name" in invalid_row["inputs"]
    assert invalid_row["inputs"]["name"] == "Test User"

    # AND the second row (valid) should have both inputs serialized
    valid_row = result.dataset[1]
    assert valid_row["label"] == "Scenario with valid PDF"
    assert "document" in valid_row["inputs"]
    assert "name" in valid_row["inputs"]
    assert valid_row["inputs"]["name"] == "Another User"



from dataclasses import asdict, is_dataclass
import inspect
import logging
from uuid import UUID
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Type, Union, cast, get_args, get_origin

from pydantic import BaseModel

from vellum.client.types.logical_operator import LogicalOperator
from vellum.utils.uuid import is_valid_uuid
from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.expressions.accessor import AccessorExpression
from vellum.workflows.expressions.add import AddExpression
from vellum.workflows.expressions.and_ import AndExpression
from vellum.workflows.expressions.begins_with import BeginsWithExpression
from vellum.workflows.expressions.between import BetweenExpression
from vellum.workflows.expressions.coalesce_expression import CoalesceExpression
from vellum.workflows.expressions.concat import ConcatExpression
from vellum.workflows.expressions.contains import ContainsExpression
from vellum.workflows.expressions.does_not_begin_with import DoesNotBeginWithExpression
from vellum.workflows.expressions.does_not_contain import DoesNotContainExpression
from vellum.workflows.expressions.does_not_end_with import DoesNotEndWithExpression
from vellum.workflows.expressions.does_not_equal import DoesNotEqualExpression
from vellum.workflows.expressions.ends_with import EndsWithExpression
from vellum.workflows.expressions.equals import EqualsExpression
from vellum.workflows.expressions.greater_than import GreaterThanExpression
from vellum.workflows.expressions.greater_than_or_equal_to import GreaterThanOrEqualToExpression
from vellum.workflows.expressions.in_ import InExpression
from vellum.workflows.expressions.is_blank import IsBlankExpression
from vellum.workflows.expressions.is_error import IsErrorExpression
from vellum.workflows.expressions.is_nil import IsNilExpression
from vellum.workflows.expressions.is_not_blank import IsNotBlankExpression
from vellum.workflows.expressions.is_not_nil import IsNotNilExpression
from vellum.workflows.expressions.is_not_null import IsNotNullExpression
from vellum.workflows.expressions.is_not_undefined import IsNotUndefinedExpression
from vellum.workflows.expressions.is_null import IsNullExpression
from vellum.workflows.expressions.is_undefined import IsUndefinedExpression
from vellum.workflows.expressions.length import LengthExpression
from vellum.workflows.expressions.less_than import LessThanExpression
from vellum.workflows.expressions.less_than_or_equal_to import LessThanOrEqualToExpression
from vellum.workflows.expressions.minus import MinusExpression
from vellum.workflows.expressions.not_between import NotBetweenExpression
from vellum.workflows.expressions.not_in import NotInExpression
from vellum.workflows.expressions.or_ import OrExpression
from vellum.workflows.expressions.parse_json import ParseJsonExpression
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.nodes.displayable.bases.utils import primitive_to_vellum_value
from vellum.workflows.nodes.utils import get_unadorned_node
from vellum.workflows.references.constant import ConstantValueReference
from vellum.workflows.references.environment_variable import EnvironmentVariableReference
from vellum.workflows.references.execution_count import ExecutionCountReference
from vellum.workflows.references.lazy import LazyReference
from vellum.workflows.references.output import OutputReference
from vellum.workflows.references.state_value import StateValueReference
from vellum.workflows.references.trigger import TriggerAttributeReference
from vellum.workflows.references.vellum_secret import VellumSecretReference
from vellum.workflows.references.workflow_input import WorkflowInputReference
from vellum.workflows.state.base import BaseState
from vellum.workflows.types.core import JsonArray, JsonObject
from vellum.workflows.types.generics import is_workflow_class
from vellum.workflows.utils.files import virtual_open
from vellum.workflows.utils.functions import compile_function_definition, compile_inline_workflow_function_definition
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum.workflows.workflows.base import BaseWorkflow
from vellum_ee.workflows.display.utils.exceptions import (
    InvalidInputReferenceError,
    InvalidOutputReferenceError,
    UnsupportedSerializationException,
)

if TYPE_CHECKING:
    from vellum_ee.workflows.display.types import WorkflowDisplayContext

logger = logging.getLogger(__name__)


def convert_descriptor_to_operator(descriptor: BaseDescriptor) -> LogicalOperator:
    if isinstance(descriptor, EqualsExpression):
        return "="
    elif isinstance(descriptor, DoesNotEqualExpression):
        return "!="
    elif isinstance(descriptor, LessThanExpression):
        return "<"
    elif isinstance(descriptor, GreaterThanExpression):
        return ">"
    elif isinstance(descriptor, LessThanOrEqualToExpression):
        return "<="
    elif isinstance(descriptor, GreaterThanOrEqualToExpression):
        return ">="
    elif isinstance(descriptor, ContainsExpression):
        return "contains"
    elif isinstance(descriptor, BeginsWithExpression):
        return "beginsWith"
    elif isinstance(descriptor, EndsWithExpression):
        return "endsWith"
    elif isinstance(descriptor, DoesNotContainExpression):
        return "doesNotContain"
    elif isinstance(descriptor, DoesNotBeginWithExpression):
        return "doesNotBeginWith"
    elif isinstance(descriptor, DoesNotEndWithExpression):
        return "doesNotEndWith"
    elif isinstance(descriptor, IsErrorExpression):
        return "isError"
    elif isinstance(descriptor, (IsNullExpression, IsNilExpression, IsUndefinedExpression)):
        return "null"
    elif isinstance(descriptor, (IsNotNullExpression, IsNotNilExpression, IsNotUndefinedExpression)):
        return "notNull"
    elif isinstance(descriptor, IsBlankExpression):
        return "blank"
    elif isinstance(descriptor, IsNotBlankExpression):
        return "notBlank"
    elif isinstance(descriptor, InExpression):
        return "in"
    elif isinstance(descriptor, NotInExpression):
        return "notIn"
    elif isinstance(descriptor, BetweenExpression):
        return "between"
    elif isinstance(descriptor, NotBetweenExpression):
        return "notBetween"
    elif isinstance(descriptor, AndExpression):
        return "and"
    elif isinstance(descriptor, OrExpression):
        return "or"
    elif isinstance(descriptor, CoalesceExpression):
        return "coalesce"
    elif isinstance(descriptor, ParseJsonExpression):
        return "parseJson"
    elif isinstance(descriptor, LengthExpression):
        return "length"
    elif isinstance(descriptor, AddExpression):
        return "+"
    elif isinstance(descriptor, MinusExpression):
        return "-"
    elif isinstance(descriptor, ConcatExpression):
        return "concat"
    else:
        raise UnsupportedSerializationException(f"Unsupported descriptor type: {descriptor}")


def get_child_descriptor(value: LazyReference, display_context: "WorkflowDisplayContext") -> BaseDescriptor:
    if isinstance(value._get, str):
        reference_parts = value._get.split(".")
        if len(reference_parts) < 3:
            raise InvalidOutputReferenceError(
                f"Failed to parse lazy reference: {value._get}. Only Node and Workflow Output references are supported."
            )

        output_name = reference_parts[-1]
        nested_class_name = reference_parts[-2]
        if nested_class_name != "Outputs":
            raise InvalidOutputReferenceError(
                f"Failed to parse lazy reference: {value._get}. Outputs are the only supported references."
            )

        reference_class_name = ".".join(reference_parts[:-2])

        # First try to find a node with the matching name
        for node in display_context.global_node_displays.keys():
            if node.__name__ == reference_class_name:
                return getattr(node.Outputs, output_name)

        # If no node found, check if it's a reference to the current workflow's outputs
        workflow_class = display_context.workflow_display_class.infer_workflow_class()
        if workflow_class.__name__ == reference_class_name:
            return getattr(workflow_class.Outputs, output_name)

        raise InvalidOutputReferenceError(
            f"Failed to parse lazy reference: {value._get}. "
            f"Could not find node or workflow class '{reference_class_name}'."
        )

    return value._get()


def _get_pydantic_model_definition(model_class: type) -> Optional[JsonObject]:
    """Extract module path definition from a Pydantic model class."""
    if not inspect.isclass(model_class) or not issubclass(model_class, BaseModel):
        return None

    module = model_class.__module__
    name = model_class.__name__

    module_path = module.split(".")

    return {
        "name": name,
        "module": cast(JsonArray, module_path),
    }


BinaryExpression = Union[
    AddExpression,
    AndExpression,
    BeginsWithExpression,
    CoalesceExpression,
    ConcatExpression,
    ContainsExpression,
    DoesNotBeginWithExpression,
    DoesNotContainExpression,
    DoesNotEndWithExpression,
    DoesNotEqualExpression,
    EndsWithExpression,
    EqualsExpression,
    GreaterThanExpression,
    GreaterThanOrEqualToExpression,
    InExpression,
    LessThanExpression,
    LessThanOrEqualToExpression,
    MinusExpression,
    NotInExpression,
    OrExpression,
]


def _serialize_condition(
    executable_id: UUID, display_context: "WorkflowDisplayContext", condition: BaseDescriptor
) -> JsonObject:
    if isinstance(
        condition,
        (
            IsBlankExpression,
            IsErrorExpression,
            IsNotBlankExpression,
            IsNullExpression,
            IsNotNullExpression,
            IsNilExpression,
            IsNotNilExpression,
            IsUndefinedExpression,
            IsNotUndefinedExpression,
            LengthExpression,
            ParseJsonExpression,
        ),
    ):
        lhs = serialize_value(executable_id, display_context, condition._expression)
        return {
            "type": "UNARY_EXPRESSION",
            "lhs": lhs,
            "operator": convert_descriptor_to_operator(condition),
        }
    elif isinstance(condition, (BetweenExpression, NotBetweenExpression)):
        base = serialize_value(executable_id, display_context, condition._value)
        lhs = serialize_value(executable_id, display_context, condition._start)
        rhs = serialize_value(executable_id, display_context, condition._end)

        return {
            "type": "TERNARY_EXPRESSION",
            "base": base,
            "operator": convert_descriptor_to_operator(condition),
            "lhs": lhs,
            "rhs": rhs,
        }
    elif isinstance(
        condition,
        get_args(BinaryExpression),
    ):
        lhs = serialize_value(executable_id, display_context, condition._lhs)
        rhs = serialize_value(executable_id, display_context, condition._rhs)

        return {
            "type": "BINARY_EXPRESSION",
            "lhs": lhs,
            "operator": convert_descriptor_to_operator(condition),
            "rhs": rhs,
        }
    elif isinstance(condition, AccessorExpression):
        return {
            "type": "BINARY_EXPRESSION",
            "lhs": serialize_value(executable_id, display_context, condition._base),
            "operator": "accessField",
            "rhs": serialize_value(executable_id, display_context, condition._field),
        }

    raise UnsupportedSerializationException(f"Unsupported condition type: {condition.__class__.__name__}")


def serialize_key(key: Any) -> str:
    """Serialize dictionary keys to strings, handling function objects properly."""
    if callable(key):
        return key.__name__
    elif isinstance(key, str):
        return key
    else:
        return str(key)


def _extract_node_classes_from_object(obj: Any, seen: Set[Type[BaseNode]], node_classes: List[Type[BaseNode]]) -> None:
    """
    Extract BaseNode subclasses from an object, handling classes, instances, and typing constructs.

    Modifies seen and node_classes in place.
    """
    if isinstance(obj, str):
        return

    if inspect.isclass(obj) and issubclass(obj, BaseNode) and obj is not BaseNode:
        if obj not in seen:
            seen.add(obj)
            node_classes.append(obj)
        return

    if isinstance(obj, BaseNode):
        obj_class = obj.__class__
        if obj_class not in seen:
            seen.add(obj_class)
            node_classes.append(obj_class)
        return

    origin = get_origin(obj)
    if origin is not None:
        for arg in get_args(obj):
            _extract_node_classes_from_object(arg, seen, node_classes)


def _get_node_references_in_callable(func: Any) -> List[Type[BaseNode]]:
    """
    Check if a callable actually references any BaseNode subclasses in its body.

    Uses inspect.getclosurevars to detect globals/nonlocals referenced by the function.

    Returns a list of node classes found in the callable's referenced scope.
    """
    node_classes: List[Type[BaseNode]] = []
    seen: Set[Type[BaseNode]] = set()

    try:
        closure_vars = inspect.getclosurevars(func)
        for obj in list(closure_vars.globals.values()) + list(closure_vars.nonlocals.values()):
            _extract_node_classes_from_object(obj, seen, node_classes)
    except (TypeError, ValueError):
        pass

    return node_classes


def serialize_value(executable_id: UUID, display_context: "WorkflowDisplayContext", value: Any) -> Optional[JsonObject]:
    """
    Serialize a value to a JSON object. Returns `None` if the value resolves to `undefined`.
    This is safe because all valid values are a JSON object, including the `None` constant:
    > `{"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}}`

    Args:
        executable_id: node id or workflow id
        display_context: workflow display context
        value: value to serialize

    Returns:
        serialized value
    """
    if value is undefined:
        return None

    if isinstance(value, ConstantValueReference):
        return serialize_value(executable_id, display_context, value._value)

    if isinstance(value, LazyReference):
        try:
            child_descriptor = get_child_descriptor(value, display_context)
        except InvalidOutputReferenceError as e:
            logger.warning("Failed to parse lazy reference '%s', skipping serialization", value.name)
            display_context.add_validation_error(e)
            return None
        return serialize_value(executable_id, display_context, child_descriptor)

    if isinstance(value, WorkflowInputReference):
        if value not in display_context.global_workflow_input_displays:
            workflow_inputs_class = display_context.workflow_display_class.infer_workflow_class().get_inputs_class()
            if value.inputs_class != workflow_inputs_class:
                inputs_class_name = value.inputs_class.__name__
                workflow_class_name = display_context.workflow_display_class.infer_workflow_class().__name__
                raise UnsupportedSerializationException(
                    f"Inputs class '{inputs_class_name}' referenced during serialization of "
                    f"'{workflow_class_name}' without parameterizing this Workflow with this Inputs definition. "
                    f"Update your Workflow definition to "
                    f"'{workflow_class_name}(BaseWorkflow[{inputs_class_name}, BaseState])'."
                )
            else:
                raise InvalidInputReferenceError(
                    message=f"Inputs class '{value.inputs_class.__qualname__}' has no attribute '{value.name}'",
                    inputs_class_name=value.inputs_class.__qualname__,
                    attribute_name=value.name,
                )
        workflow_input_display = display_context.global_workflow_input_displays[value]
        return {
            "type": "WORKFLOW_INPUT",
            "input_variable_id": str(workflow_input_display.id),
        }

    if isinstance(value, StateValueReference):
        state_value_display = display_context.global_state_value_displays[value]
        return {
            "type": "WORKFLOW_STATE",
            "state_variable_id": str(state_value_display.id),
        }

    if isinstance(value, OutputReference):
        if issubclass(value.outputs_class, BaseWorkflow.Outputs):
            if value in display_context.workflow_output_displays:
                workflow_output_display = display_context.workflow_output_displays[value]
                return {
                    "type": "WORKFLOW_OUTPUT",
                    "output_variable_id": str(workflow_output_display.id),
                }
            return serialize_value(executable_id, display_context, value.instance)

        if value not in display_context.global_node_output_displays:
            if issubclass(value.outputs_class, BaseNode.Outputs):
                raise InvalidOutputReferenceError(
                    f"Reference to node '{value.outputs_class.__parent_class__.__name__}' not found in graph."
                )
            raise InvalidOutputReferenceError(f"Reference to outputs '{value.outputs_class.__qualname__}' is invalid.")

        output_display = display_context.global_node_output_displays[value]

        upstream_node_class = value.outputs_class.__parent_class__
        if not issubclass(upstream_node_class, BaseNode):
            raise ValueError(f"Output references must be to a node, not {upstream_node_class}")
        unadorned_upstream_node_class = get_unadorned_node(upstream_node_class)
        upstream_node = display_context.global_node_displays[unadorned_upstream_node_class]

        return {
            "type": "NODE_OUTPUT",
            "node_id": str(upstream_node.node_id),
            "node_output_id": str(output_display.id),
        }

    if isinstance(value, VellumSecretReference):
        return {
            "type": "VELLUM_SECRET",
            "vellum_secret_name": value.name,
        }

    if isinstance(value, EnvironmentVariableReference):
        return {
            "type": "ENVIRONMENT_VARIABLE",
            "environment_variable": value.name,
        }

    if isinstance(value, ExecutionCountReference):
        node_class_display = display_context.global_node_displays[value.node_class]

        return {
            "type": "EXECUTION_COUNTER",
            "node_id": str(node_class_display.node_id),
        }

    if isinstance(value, TriggerAttributeReference):
        trigger_class = value.trigger_class
        trigger_id = trigger_class.__id__

        return {
            "type": "TRIGGER_ATTRIBUTE",
            "trigger_id": str(trigger_id),
            "attribute_id": str(value.id),
        }

    if isinstance(value, list):
        serialized_items = []
        for item in value:
            serialized_item = serialize_value(executable_id, display_context, item)
            if serialized_item is not None:
                serialized_items.append(serialized_item)

        if all(isinstance(item, dict) and item["type"] == "CONSTANT_VALUE" for item in serialized_items):
            constant_values = []
            for item in serialized_items:
                item_dict = cast(Dict[str, Any], item)
                value_inner = item_dict["value"]

                if value_inner["type"] == "JSON" and "items" in value_inner:
                    # Nested JSON list
                    constant_values.append(value_inner["items"])
                else:
                    constant_values.append(value_inner["value"])

            return {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "JSON",
                    "value": constant_values,
                },
            }
        else:
            return {
                "type": "ARRAY_REFERENCE",
                "items": cast(JsonArray, serialized_items),  # list[JsonObject] -> JsonArray
            }

    if is_dataclass(value) and not isinstance(value, type):
        dict_value = asdict(value)
        return serialize_value(executable_id, display_context, dict_value)

    if isinstance(value, dict):
        serialized_entries: List[Dict[str, Any]] = []
        for key, val in value.items():
            serialized_val = serialize_value(executable_id, display_context, val)
            if serialized_val is not None:
                serialized_entries.append(
                    {
                        "id": str(uuid4_from_hash(f"{executable_id}|{key}")),
                        "key": serialize_key(key),
                        "value": serialized_val,
                    }
                )

        # Check if all entries have constant values
        if all(entry["value"]["type"] == "CONSTANT_VALUE" for entry in serialized_entries):
            constant_entries = {}
            for entry in serialized_entries:
                entry_value = entry["value"]["value"]
                constant_entries[entry["key"]] = entry_value["value"]

            return {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "JSON",
                    "value": constant_entries,
                },
            }
        else:
            return {
                "type": "DICTIONARY_REFERENCE",
                "entries": cast(JsonArray, serialized_entries),
            }

    if is_workflow_class(value):
        # Import here to avoid circular imports
        from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

        # Pass the parent display context so the subworkflow can resolve parent workflow inputs
        workflow_display = get_workflow_display(
            workflow_class=value,
            parent_display_context=display_context,
        )
        serialized_value: dict = workflow_display.serialize()
        name = serialized_value["workflow_raw_data"]["definition"]["name"]
        description = value.__doc__ or ""

        # Compile the function definition (which now handles __vellum_inputs__ and __vellum_examples__)
        function_definition = compile_inline_workflow_function_definition(value)

        # Handle __vellum_inputs__ for inline workflows (similar to function tools)
        inputs = getattr(value, "__vellum_inputs__", {})
        if inputs:
            serialized_inputs = {}
            for param_name, input_ref in inputs.items():
                serialized_input = serialize_value(executable_id, display_context, input_ref)
                if serialized_input is not None:
                    serialized_inputs[param_name] = serialized_input

            model_data = function_definition.model_dump()
            model_data["inputs"] = serialized_inputs
            function_definition_data = model_data
        else:
            function_definition_data = function_definition.model_dump()

        workflow_data: Dict[str, Any] = {
            "type": "INLINE_WORKFLOW",
            "name": name,
            "description": description,
            "definition": function_definition_data,
            "exec_config": serialized_value,
        }

        return {
            "type": "CONSTANT_VALUE",
            "value": {
                "type": "JSON",
                "value": workflow_data,
            },
        }

    if isinstance(value, BaseModel):
        context = {"executable_id": executable_id, "client": display_context.client}
        dict_value = value.model_dump(context=context)
        dict_ref = serialize_value(executable_id, display_context, dict_value)
        if dict_ref is not None and dict_ref.get("type") == "DICTIONARY_REFERENCE":
            dict_ref["definition"] = _get_pydantic_model_definition(value.__class__)
        return dict_ref

    if callable(value):
        node_references = _get_node_references_in_callable(value)
        if node_references:
            node_names = ", ".join(node.__name__ for node in node_references)
            raise UnsupportedSerializationException(
                f"Code tools cannot reference workflow nodes ({node_names}). "
                "Consider inlining the node's run method logic directly in your function, "
                "or use an Inline Subworkflow tool instead."
            )

        function_definition = compile_function_definition(value)

        name = function_definition.name
        description = function_definition.description or ""

        inputs = getattr(value, "__vellum_inputs__", {})

        if inputs:
            serialized_inputs = {}
            for param_name, input_ref in inputs.items():
                serialized_input = serialize_value(executable_id, display_context, input_ref)
                if serialized_input is not None:
                    serialized_inputs[param_name] = serialized_input

            model_data = function_definition.model_dump()
            model_data["inputs"] = serialized_inputs
            function_definition_data = model_data
        else:
            function_definition_data = function_definition.model_dump()

        source_path = inspect.getsourcefile(value)
        if source_path is not None:
            with virtual_open(source_path) as f:
                source_code = f.read()
        else:
            source_code = f"Source code not available for {value.__name__}"

        return {
            "type": "CONSTANT_VALUE",
            "value": {
                "type": "JSON",
                "value": {
                    "type": "CODE_EXECUTION",
                    "name": name,
                    "description": description,
                    "definition": function_definition_data,
                    "src": source_code,
                },
            },
        }

    if not isinstance(value, BaseDescriptor):
        vellum_value = primitive_to_vellum_value(value)
        return {
            "type": "CONSTANT_VALUE",
            "value": vellum_value.dict(),
        }

    # If it's not any of the references we know about,
    # then try to serialize it as a nested value
    return _serialize_condition(executable_id, display_context, value)


def base_descriptor_validator(value: dict, workflow: Type[BaseWorkflow]) -> BaseDescriptor:
    descriptor_type = value.get("type")

    if descriptor_type == "CONSTANT_VALUE":
        return _constant_value_reference_validator(value.get("value"))

    if descriptor_type == "WORKFLOW_INPUT":
        return _workflow_input_reference_validator(value.get("input_variable_id"), workflow)

    if descriptor_type == "WORKFLOW_STATE":
        return _workflow_state_reference_validator(value.get("state_variable_id"), workflow)

    if descriptor_type == "NODE_OUTPUT":
        return _node_output_reference_validator(value.get("node_id"), value.get("node_output_id"), workflow)

    if descriptor_type == "EXECUTION_COUNTER":
        return _execution_counter_reference_validator(value.get("node_id"), workflow)

    if descriptor_type == "VELLUM_SECRET":
        return _vellum_secret_reference_validator(value.get("vellum_secret_name"))

    if descriptor_type == "ENVIRONMENT_VARIABLE":
        return _environment_variable_reference_validator(value.get("environment_variable"))

    if descriptor_type == "ARRAY_REFERENCE":
        return _array_reference_validator(value.get("items"), workflow)

    if descriptor_type == "DICTIONARY_REFERENCE":
        return _dictionary_reference_validator(value.get("entries"), workflow)

    if descriptor_type == "BINARY_EXPRESSION":
        return _binary_expression_validator(value, workflow)

    if descriptor_type == "UNARY_EXPRESSION":
        return _unary_expression_validator(value, workflow)

    if descriptor_type == "TERNARY_EXPRESSION":
        return _ternary_expression_validator(value, workflow)

    raise ValueError(f"Unsupported descriptor type: {descriptor_type}")


def _constant_value_reference_validator(constant: Any) -> ConstantValueReference:
    if not isinstance(constant, dict):
        raise ValueError(f"Unexpected type for constant reference: {type(constant)}")

    return ConstantValueReference(constant.get("value"))


def _workflow_input_reference_validator(workflow_input_id: Any, workflow: Type[BaseWorkflow]) -> WorkflowInputReference:
    inputs_class = workflow.get_inputs_class()
    if not issubclass(inputs_class, BaseInputs):
        raise ValueError(f"Unexpected type for inputs class: {type(inputs_class)}")

    input_reference = next(
        (
            input
            for input in inputs_class
            if isinstance(input, WorkflowInputReference) and str(input.id) == str(workflow_input_id)
        ),
        None,
    )
    if input_reference is None:
        raise ValueError(f"Input reference not found: {workflow_input_id}")

    return input_reference


def _workflow_state_reference_validator(workflow_state_id: Any, workflow: Type[BaseWorkflow]) -> StateValueReference:
    state_class = workflow.get_state_class()
    if not issubclass(state_class, BaseState):
        raise ValueError(f"Unexpected type for state class: {type(state_class)}")

    state_reference = next((state for state in state_class if str(state.id) == str(workflow_state_id)), None)
    if state_reference is None:
        raise ValueError(f"State reference not found: {workflow_state_id}")

    return state_reference


def _node_output_reference_validator(
    node_id: Any, node_output_id: Any, workflow: Type[BaseWorkflow]
) -> OutputReference:
    if not is_valid_uuid(node_id):
        raise ValueError(f"Unexpected type for node id: {type(node_id)}")

    if not is_valid_uuid(node_output_id):
        raise ValueError(f"Unexpected type for node output id: {type(node_output_id)}")

    node_class = workflow.resolve_node_ref(node_id)
    if not issubclass(node_class, BaseNode):
        raise ValueError(f"Unexpected type for node class: {type(node_class)}")

    output_reference = next((output for output in node_class.Outputs if str(output.id) == str(node_output_id)), None)
    if output_reference is None:
        raise ValueError(f"Output reference not found: {node_output_id}")

    return output_reference


def _execution_counter_reference_validator(node_id: Any, workflow: Type[BaseWorkflow]) -> ExecutionCountReference:
    node_class = workflow.resolve_node_ref(node_id)
    return ExecutionCountReference(node_class)


def _vellum_secret_reference_validator(secret_name: Any) -> VellumSecretReference:
    if not isinstance(secret_name, str):
        raise ValueError(f"Unexpected type for vellum secret name: {type(secret_name)}")
    return VellumSecretReference(secret_name)


def _environment_variable_reference_validator(env_var_name: Any) -> EnvironmentVariableReference:
    if not isinstance(env_var_name, str):
        raise ValueError(f"Unexpected type for environment variable name: {type(env_var_name)}")
    return EnvironmentVariableReference(name=env_var_name)


def _array_reference_validator(items: Any, workflow: Type[BaseWorkflow]) -> ConstantValueReference:
    if not isinstance(items, list):
        raise ValueError(f"Unexpected type for array items: {type(items)}")

    validated_items = []
    for item in items:
        if isinstance(item, dict):
            validated_items.append(base_descriptor_validator(item, workflow))
        else:
            validated_items.append(item)

    return ConstantValueReference(validated_items)


def _dictionary_reference_validator(entries: Any, workflow: Type[BaseWorkflow]) -> ConstantValueReference:
    if not isinstance(entries, list):
        raise ValueError(f"Unexpected type for dictionary entries: {type(entries)}")

    validated_dict: Dict[str, Any] = {}
    for entry in entries:
        if not isinstance(entry, dict):
            raise ValueError(f"Unexpected type for dictionary entry: {type(entry)}")

        key = entry.get("key")
        if not isinstance(key, str):
            raise ValueError(f"Unexpected type for dictionary key: {type(key)}")

        raw_value = entry.get("value")
        if isinstance(raw_value, dict):
            validated_dict[key] = base_descriptor_validator(raw_value, workflow)
        else:
            validated_dict[key] = raw_value

    return ConstantValueReference(validated_dict)


UnaryExpression = Union[
    IsBlankExpression,
    IsErrorExpression,
    IsNilExpression,
    IsNotBlankExpression,
    IsNotNilExpression,
    IsNotNullExpression,
    IsNotUndefinedExpression,
    IsNullExpression,
    IsUndefinedExpression,
    LengthExpression,
    ParseJsonExpression,
]


def _unary_expression_validator(unary_expression: dict, workflow: Type[BaseWorkflow]) -> UnaryExpression:
    operator = unary_expression.get("operator")
    raw_lhs = unary_expression.get("lhs")
    if not isinstance(raw_lhs, dict):
        raise ValueError(f"Unexpected type for lhs: {type(raw_lhs)}")

    expression = base_descriptor_validator(raw_lhs, workflow)

    if operator == "blank":
        return IsBlankExpression(expression=expression)
    elif operator == "notBlank":
        return IsNotBlankExpression(expression=expression)
    elif operator == "null":
        return IsNullExpression(expression=expression)
    elif operator == "notNull":
        return IsNotNullExpression(expression=expression)
    elif operator == "isError":
        return IsErrorExpression(expression=expression)
    elif operator == "length":
        return LengthExpression(expression=expression)
    elif operator == "parseJson":
        return ParseJsonExpression(expression=expression)

    raise ValueError(f"Unsupported unary expression operator: {operator}")


TernaryExpression = Union[BetweenExpression, NotBetweenExpression]


def _ternary_expression_validator(ternary_expression: dict, workflow: Type[BaseWorkflow]) -> TernaryExpression:
    operator = ternary_expression.get("operator")
    raw_base = ternary_expression.get("base")
    if not isinstance(raw_base, dict):
        raise ValueError(f"Unexpected type for base: {type(raw_base)}")

    raw_lhs = ternary_expression.get("lhs")
    if not isinstance(raw_lhs, dict):
        raise ValueError(f"Unexpected type for lhs: {type(raw_lhs)}")

    raw_rhs = ternary_expression.get("rhs")
    if not isinstance(raw_rhs, dict):
        raise ValueError(f"Unexpected type for rhs: {type(raw_rhs)}")

    base = base_descriptor_validator(raw_base, workflow)
    start = base_descriptor_validator(raw_lhs, workflow)
    end = base_descriptor_validator(raw_rhs, workflow)

    if operator == "between":
        return BetweenExpression(value=base, start=start, end=end)
    elif operator == "notBetween":
        return NotBetweenExpression(value=base, start=start, end=end)

    raise ValueError(f"Unsupported ternary expression operator: {operator}")


def _binary_expression_validator(
    binary_expression: dict, workflow: Type[BaseWorkflow]
) -> Union[BinaryExpression, AccessorExpression]:
    operator = binary_expression.get("operator")
    raw_lhs = binary_expression.get("lhs")
    if not isinstance(raw_lhs, dict):
        raise ValueError(f"Unexpected type for lhs: {type(raw_lhs)}")

    raw_rhs = binary_expression.get("rhs")
    if not isinstance(raw_rhs, dict):
        raise ValueError(f"Unexpected type for rhs: {type(raw_rhs)}")

    lhs = base_descriptor_validator(raw_lhs, workflow)
    rhs = base_descriptor_validator(raw_rhs, workflow)

    if operator in ("=", "=="):
        return EqualsExpression(lhs=lhs, rhs=rhs)
    elif operator == "!=":
        return DoesNotEqualExpression(lhs=lhs, rhs=rhs)
    elif operator == "<":
        return LessThanExpression(lhs=lhs, rhs=rhs)
    elif operator == ">":
        return GreaterThanExpression(lhs=lhs, rhs=rhs)
    elif operator == "<=":
        return LessThanOrEqualToExpression(lhs=lhs, rhs=rhs)
    elif operator == ">=":
        return GreaterThanOrEqualToExpression(lhs=lhs, rhs=rhs)
    elif operator == "contains":
        return ContainsExpression(lhs=lhs, rhs=rhs)
    elif operator == "beginsWith":
        return BeginsWithExpression(lhs=lhs, rhs=rhs)
    elif operator == "endsWith":
        return EndsWithExpression(lhs=lhs, rhs=rhs)
    elif operator == "doesNotContain":
        return DoesNotContainExpression(lhs=lhs, rhs=rhs)
    elif operator == "doesNotBeginWith":
        return DoesNotBeginWithExpression(lhs=lhs, rhs=rhs)
    elif operator == "doesNotEndWith":
        return DoesNotEndWithExpression(lhs=lhs, rhs=rhs)
    elif operator == "in":
        return InExpression(lhs=lhs, rhs=rhs)
    elif operator == "notIn":
        return NotInExpression(lhs=lhs, rhs=rhs)
    elif operator == "and":
        return AndExpression(lhs=lhs, rhs=rhs)
    elif operator == "or":
        return OrExpression(lhs=lhs, rhs=rhs)
    elif operator == "coalesce":
        return CoalesceExpression(lhs=lhs, rhs=rhs)
    elif operator == "+":
        return AddExpression(lhs=lhs, rhs=rhs)
    elif operator == "-":
        return MinusExpression(lhs=lhs, rhs=rhs)
    elif operator == "concat":
        return ConcatExpression(lhs=lhs, rhs=rhs)
    elif operator == "accessField":
        return AccessorExpression(base=lhs, field=rhs)

    raise ValueError(f"Unsupported binary expression operator: {operator}")



from abc import ABC, ABCMeta
from functools import lru_cache
import inspect
import json
import os
from uuid import UUID
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Iterator, Optional, Tuple, Type, cast, get_origin

from vellum.workflows.references.trigger import TriggerAttributeReference
from vellum.workflows.types.utils import get_class_attr_names, infer_types
from vellum.workflows.utils.files import virtual_open
from vellum.workflows.utils.uuids import get_trigger_attribute_id, uuid4_from_hash
from vellum_ee.workflows.display.editor import NodeDisplayComment

if TYPE_CHECKING:
    from vellum.workflows.graph.graph import Graph, GraphTarget
    from vellum.workflows.state.base import BaseState


def _is_annotated(cls: Type, name: str) -> bool:
    annotations = getattr(cls, "__annotations__", {})
    annotation = annotations.get(name)
    if annotation is not None:
        if get_origin(annotation) is ClassVar:
            return False
        return True

    for base in cls.__bases__:
        if _is_annotated(base, name):
            return True

    return False


def _convert_to_relative_module_path(absolute_module_path: str, workflow_root: str) -> str:
    """
    Convert an absolute module path to a relative path from the workflow root.

    Args:
        absolute_module_path: The full module path (e.g., "workflow_id.triggers.scheduled")
        workflow_root: The workflow root module (e.g., "workflow_id")

    Returns:
        Relative module path with leading dot (e.g., ".triggers.scheduled")
    """
    if not absolute_module_path.startswith(workflow_root):
        return ""

    remaining_path = absolute_module_path[len(workflow_root) :]
    if remaining_path.startswith("."):
        return remaining_path
    else:
        return "." + remaining_path


def _find_workflow_root_with_metadata(trigger_module: str) -> Optional[str]:
    """
    Find the workflow root module by searching for metadata.json up the module hierarchy.

    Args:
        trigger_module: The trigger's module path (e.g., "workflows.my_workflow.triggers.my_trigger")

    Returns:
        The workflow root module path if found, None otherwise
    """
    module_parts = trigger_module.split(".")

    # Try searching up the module hierarchy for metadata.json
    for i in range(len(module_parts), 0, -1):
        potential_root = ".".join(module_parts[:i])
        module_dir = potential_root.replace(".", os.path.sep)
        metadata_path = os.path.join(module_dir, "metadata.json")

        # Try to open the file using virtual_open to support both regular and virtual filesystems
        # virtual_open checks BaseWorkflowFinder instances before falling back to regular open()
        try:
            file_handle = virtual_open(metadata_path)
            if file_handle is not None:
                file_handle.close()
                return potential_root
        except (FileNotFoundError, OSError):
            pass

    return None


def _get_trigger_id_from_metadata(trigger_class: Type["BaseTrigger"]) -> Optional[UUID]:
    """
    Get the trigger ID from metadata.json for a given trigger class.

    Args:
        trigger_class: The trigger class to look up

    Returns:
        The UUID from metadata.json, or None if not found
    """
    workflow_root = _find_workflow_root_with_metadata(trigger_class.__module__)
    if not workflow_root:
        return None

    trigger_path_to_id_mapping = _get_trigger_path_to_id_mapping(trigger_class.__module__)
    if not trigger_path_to_id_mapping:
        return None

    # Convert module path to relative path and append class name
    # e.g., "root_module.triggers.scheduled" + "ScheduleTrigger" -> ".triggers.scheduled.ScheduleTrigger"
    relative_module_path = _convert_to_relative_module_path(trigger_class.__module__, workflow_root)
    if not relative_module_path:
        return None

    # Append class name to get full trigger path
    relative_trigger_path = f"{relative_module_path}.{trigger_class.__qualname__}"
    return trigger_path_to_id_mapping.get(relative_trigger_path)


@lru_cache(maxsize=128)
def _get_trigger_path_to_id_mapping(module_path: str) -> Dict[str, UUID]:
    """
    Read trigger path to ID mapping from metadata.json for a given module.

    This function is cached to avoid repeated file reads. It searches up the module
    hierarchy for metadata.json and extracts the trigger_path_to_id_mapping.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow.triggers.my_trigger")

    Returns:
        Dictionary mapping trigger module paths to their UUIDs
    """
    try:
        # Find the workflow root that contains metadata.json
        workflow_root = _find_workflow_root_with_metadata(module_path)
        if not workflow_root:
            return {}

        # Convert module path to file path
        module_dir = workflow_root.replace(".", os.path.sep)
        metadata_path = os.path.join(module_dir, "metadata.json")

        # Use virtual_open to support both regular and virtual environments
        with virtual_open(metadata_path) as f:
            metadata_json = json.load(f)
            trigger_mapping = metadata_json.get("trigger_path_to_id_mapping", {})

            # Convert string IDs to UUIDs
            return {path: UUID(id_str) for path, id_str in trigger_mapping.items()}

    except (FileNotFoundError, json.JSONDecodeError, ValueError, KeyError):
        # If there's any error reading or parsing the file, return empty dict
        return {}


@lru_cache(maxsize=128)
def _get_trigger_attribute_id_mapping(module_path: str) -> Dict[str, UUID]:
    """
    Read trigger attribute ID mapping from metadata.json for a given module.

    This function is cached to avoid repeated file reads. It searches up the module
    hierarchy for metadata.json and extracts the trigger_attribute_id_mapping.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow.triggers.my_trigger")

    Returns:
        Dictionary mapping "<trigger_path>|<attribute_key>" to their UUIDs
    """
    try:
        # Find the workflow root that contains metadata.json
        workflow_root = _find_workflow_root_with_metadata(module_path)
        if not workflow_root:
            return {}

        # Convert module path to file path
        module_dir = workflow_root.replace(".", os.path.sep)
        metadata_path = os.path.join(module_dir, "metadata.json")

        # Use virtual_open to support both regular and virtual environments
        with virtual_open(metadata_path) as f:
            metadata_json = json.load(f)
            raw_mapping = metadata_json.get("trigger_attribute_id_mapping", {}) or {}

        # Convert string IDs to UUIDs, skipping invalid entries
        result: Dict[str, UUID] = {}
        for key, id_str in raw_mapping.items():
            try:
                result[key] = UUID(id_str)
            except ValueError:
                # Skip invalid UUID entries; they will fall back to hash-based IDs
                continue

        return result

    except (FileNotFoundError, json.JSONDecodeError, KeyError):
        # If there's any error reading or parsing the file, return empty dict
        return {}


def _get_trigger_attribute_id_from_metadata(trigger_class: Type["BaseTrigger"], attribute_name: str) -> Optional[UUID]:
    """
    Get the trigger attribute ID from metadata.json for a given trigger class and attribute name.

    Args:
        trigger_class: The trigger class containing the attribute
        attribute_name: The name of the attribute

    Returns:
        The UUID from metadata.json, or None if not found
    """
    workflow_root = _find_workflow_root_with_metadata(trigger_class.__module__)
    if not workflow_root:
        return None

    attribute_mapping = _get_trigger_attribute_id_mapping(trigger_class.__module__)
    if not attribute_mapping:
        return None

    # Convert module path to relative path and append class name
    # e.g., "root_module.triggers.scheduled" + "ScheduleTrigger" -> ".triggers.scheduled.ScheduleTrigger"
    relative_module_path = _convert_to_relative_module_path(trigger_class.__module__, workflow_root)
    if not relative_module_path:
        return None

    # Build the key: "<trigger_path>|<attribute_key>"
    relative_trigger_path = f"{relative_module_path}.{trigger_class.__qualname__}"
    key = f"{relative_trigger_path}|{attribute_name}"
    return attribute_mapping.get(key)


class BaseTriggerMeta(ABCMeta):
    def __new__(mcs, name: str, bases: Tuple[Type, ...], dct: Dict[str, Any]) -> Any:
        if "Display" not in dct:
            for base in reversed(bases):
                if hasattr(base, "Display"):
                    dct["Display"] = type(
                        f"{name}.Display",
                        (base.Display,),
                        {"__module__": dct["__module__"]},
                    )
                    break

        cls = super().__new__(mcs, name, bases, dct)
        trigger_class = cast(Type["BaseTrigger"], cls)

        # Set default ID based on module and class name
        trigger_class.__id__ = uuid4_from_hash(f"{trigger_class.__module__}.{trigger_class.__qualname__}")

        # Try to override with ID from metadata.json if available
        metadata_id = _get_trigger_id_from_metadata(trigger_class)
        if metadata_id:
            trigger_class.__id__ = metadata_id

        return trigger_class

    """
    Metaclass for BaseTrigger that enables class-level >> operator.

    This allows triggers to be used at the class level, similar to nodes:
        ManualTrigger >> MyNode  # Class-level, no instantiation
    """

    def __getattribute__(cls, name: str) -> Any:
        trigger_cls = cast(Type["BaseTrigger"], cls)
        if name.startswith("_"):
            return super().__getattribute__(name)

        try:
            attribute = super().__getattribute__(name)
        except AttributeError as exc:
            if _is_annotated(cls, name):
                attribute = None
            else:
                raise exc

        if inspect.isroutine(attribute) or isinstance(attribute, (property, classmethod, staticmethod)):
            return attribute

        if isinstance(attribute, TriggerAttributeReference):
            return attribute

        if name in cls.__dict__:
            return attribute

        if not _is_annotated(cls, name):
            return attribute

        types = infer_types(cls, name)
        reference = TriggerAttributeReference(name=name, types=types, instance=attribute, trigger_class=trigger_cls)
        return reference

    def __iter__(cls) -> Iterator[TriggerAttributeReference]:
        seen: Dict[str, TriggerAttributeReference] = {}

        for attr_name in get_class_attr_names(cls):
            if attr_name in seen:
                yield seen[attr_name]
                continue

            # Skip ClassVar annotations - they're not trigger attributes
            if not _is_annotated(cls, attr_name):
                continue

            attr_value = getattr(cls, attr_name)
            if isinstance(attr_value, TriggerAttributeReference):
                seen[attr_name] = attr_value
                yield attr_value

    def __rshift__(cls, other: "GraphTarget") -> "Graph":  # type: ignore[misc]
        """
        Enable Trigger class >> Node syntax (class-level only).

        Args:
            other: The target to connect to - can be a Node, Graph, or set of Nodes

        Returns:
            Graph: A graph object with trigger edges

        Examples:
            ManualTrigger >> MyNode
            ManualTrigger >> {NodeA, NodeB}
            ManualTrigger >> (NodeA >> NodeB)
        """
        from vellum.workflows.edges.trigger_edge import TriggerEdge
        from vellum.workflows.graph.graph import Graph
        from vellum.workflows.nodes.bases.base import BaseNode as BaseNodeClass

        # Cast cls to the proper type for TriggerEdge
        trigger_cls = cast(Type["BaseTrigger"], cls)

        if isinstance(other, set):
            # Trigger >> {NodeA, NodeB}
            trigger_edges = []
            graph_items = []
            for item in other:
                if isinstance(item, type) and issubclass(item, BaseNodeClass):
                    trigger_edges.append(TriggerEdge(trigger_cls, item))
                elif isinstance(item, Graph):
                    # Trigger >> {Graph1, Graph2}
                    graph_items.append(item)
                    for entrypoint in item.entrypoints:
                        trigger_edges.append(TriggerEdge(trigger_cls, entrypoint))
                else:
                    raise TypeError(
                        f"Cannot connect trigger to {type(item).__name__}. " f"Expected BaseNode or Graph in set."
                    )

            result_graph = Graph.from_trigger_edges(trigger_edges)

            for graph_item in graph_items:
                result_graph._extend_edges(graph_item.edges)
                result_graph._terminals.update(graph_item._terminals)
                for existing_trigger_edge in graph_item._trigger_edges:
                    if existing_trigger_edge not in result_graph._trigger_edges:
                        result_graph._trigger_edges.append(existing_trigger_edge)

            return result_graph

        elif isinstance(other, Graph):
            # Trigger >> Graph
            edges = [TriggerEdge(trigger_cls, entrypoint) for entrypoint in other.entrypoints]
            result_graph = Graph.from_trigger_edges(edges)
            # Also include the edges from the original graph
            result_graph._extend_edges(other.edges)
            result_graph._terminals = other._terminals
            return result_graph

        elif isinstance(other, type) and issubclass(other, BaseNodeClass):
            # Trigger >> Node
            edge = TriggerEdge(trigger_cls, other)
            return Graph.from_trigger_edge(edge)

        else:
            raise TypeError(
                f"Cannot connect trigger to {type(other).__name__}. " f"Expected BaseNode, Graph, or set of BaseNodes."
            )

    def __rrshift__(cls, other: Any) -> "Graph":
        """
        Prevent Node >> Trigger class syntax.

        Raises:
            TypeError: Always, as this operation is not allowed
        """
        raise TypeError(
            f"Cannot create edge targeting trigger {cls.__name__}. "
            f"Triggers must be at the start of a graph path, not as targets. "
            f"Did you mean: {cls.__name__} >> {other.__name__ if hasattr(other, '__name__') else other}?"
        )


class BaseTrigger(ABC, metaclass=BaseTriggerMeta):
    """
    Base class for workflow triggers - first-class graph elements.

    Triggers define how and when a workflow execution is initiated. They are integrated
    into the workflow graph using the >> operator and can connect to nodes at the class level.

    Examples:
        # Class-level usage (consistent with nodes)
        ManualTrigger >> MyNode
        ManualTrigger >> {NodeA, NodeB}
        ManualTrigger >> (NodeA >> NodeB)

    Subclass Hierarchy:
        - ManualTrigger: Explicit workflow invocation (default)
        - IntegrationTrigger: External service triggers (base for Slack, GitHub, etc.)
        - ScheduledTrigger: Time-based triggers with cron/interval schedules

    Important:
        Triggers can only appear at the start of graph paths. Attempting to create
        edges targeting triggers (Node >> Trigger) will raise a TypeError.

    Note:
        Like nodes, triggers work at the class level only. Do not instantiate triggers.
    """

    __id__: UUID

    class Display:
        """Optional display metadata for visual representation."""

        label: str = "Trigger"
        x: float = 0.0
        y: float = 0.0
        z_index: int = 0
        icon: Optional[str] = None
        color: Optional[str] = None
        comment: Optional[NodeDisplayComment] = None

    def __init__(self, **kwargs: Any):
        """
        Initialize trigger with event data.

        Args:
            **kwargs: Arbitrary keyword arguments passed during trigger instantiation.
                     Subclasses may use these to populate trigger attributes.
        """
        self._event_data = kwargs

        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def attribute_references(cls) -> Dict[str, "TriggerAttributeReference[Any]"]:
        """Return class-level trigger attribute descriptors keyed by attribute name."""

        return {reference.name: reference for reference in cls}

    def to_trigger_attribute_values(self) -> Dict["TriggerAttributeReference[Any]", Any]:
        """Materialize attribute descriptor/value pairs for this trigger instance."""

        attribute_values: Dict["TriggerAttributeReference[Any]", Any] = {}
        for reference in type(self):
            if hasattr(self, reference.name):
                attribute_values[reference] = getattr(self, reference.name)
            elif type(None) in reference.types:
                attribute_values[reference] = None
            else:
                raise AttributeError(
                    f"{type(self).__name__} did not populate required trigger attribute '{reference.name}'"
                )

        return attribute_values

    def bind_to_state(self, state: "BaseState") -> None:
        """Persist this trigger's attribute values onto the provided state."""

        if state.meta.trigger_attributes is None:
            state.meta.trigger_attributes = {}
        state.meta.trigger_attributes.update(self.to_trigger_attribute_values())

    @classmethod
    def get_attribute_id(cls, attribute_name: str) -> UUID:
        """
        Get the ID for a trigger attribute, first checking metadata.json for a stable ID,
        then falling back to a deterministic hash-based ID.

        This ensures trigger attribute IDs remain stable across serialization round-trips
        when metadata.json is present.

        Args:
            attribute_name: The name of the attribute

        Returns:
            The UUID for the attribute
        """
        # First try to get the ID from metadata.json
        metadata_id = _get_trigger_attribute_id_from_metadata(cls, attribute_name)
        if metadata_id is not None:
            return metadata_id

        # Fall back to deterministic hash-based ID
        return get_trigger_attribute_id(cls, attribute_name)

    class Config:
        """Configuration for trigger behavior. Subclasses can override."""

        pass

    def __on_workflow_initiated__(self, state: "BaseState") -> None:
        """
        Lifecycle hook called by WorkflowRunner when workflow execution starts.

        Subclasses can override to perform state initialization (e.g., appending
        user messages to chat history so nodes can reference them).

        Args:
            state: The initial workflow state
        """
        pass

    def __on_workflow_fulfilled__(self, state: "BaseState") -> None:
        """
        Lifecycle hook called by WorkflowRunner before workflow outputs are resolved.

        Subclasses can override to perform state updates that should be reflected
        in workflow outputs (e.g., appending assistant messages to chat history).

        Args:
            state: The final workflow state
        """
        pass



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class UpdateActiveWorkspaceResponse(UniversalBaseModel):
    update_active_workspace_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The id of the workspace that the user should update to, or null if no workspace change needed.
    """

    update_active_environment_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The id of the environment that the user should update to, or null if no environment change needed.
    """

    is_staff: typing.Optional[bool] = pydantic.Field(default=None)
    """
    Whether or not the user is a staff member of Vellum.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from datetime import datetime
from uuid import uuid4

from deepdiff import DeepDiff

from vellum import (
    ReleaseEnvironment,
    VellumVariable,
    WorkflowDeploymentRelease,
    WorkflowDeploymentReleaseWorkflowDeployment,
    WorkflowDeploymentReleaseWorkflowVersion,
)
from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import SubworkflowDeploymentNode, TemplatingNode
from vellum.workflows.nodes.core.try_node.node import TryNode
from vellum.workflows.outputs import BaseOutputs
from vellum.workflows.state import BaseState
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_subworkflow_deployment.workflow import BasicSubworkflowDeploymentWorkflow


def test_serialize_workflow(vellum_client):
    # GIVEN a Workflow with stubbed out API calls
    deployment_id = str(uuid4())
    deployment_release = WorkflowDeploymentRelease(
        id=str(uuid4()),
        created=datetime.now(),
        environment=ReleaseEnvironment(id=str(uuid4()), name="DEVELOPMENT", label="Development"),
        workflow_version=WorkflowDeploymentReleaseWorkflowVersion(
            id=str(uuid4()),
            input_variables=[],
            output_variables=[],
        ),
        deployment=WorkflowDeploymentReleaseWorkflowDeployment(
            id=deployment_id,
            name="example_subworkflow_deployment",
        ),
        release_tags=[],
        reviews=[],
    )
    vellum_client.workflow_deployments.retrieve_workflow_deployment_release.return_value = deployment_release

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicSubworkflowDeploymentWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "693cc9a5-8d74-4a58-bdcf-2b4989cdf250",
                "key": "city",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
                "schema": {"type": "string"},
            },
            {
                "id": "19a78824-9a98-4ae8-a1fc-61f81a422a17",
                "key": "date",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
                "schema": {"type": "string"},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "3f487916-126f-4d6c-95b4-fa72d875b793",
                "key": "temperature",
                "type": "NUMBER",
            },
            {
                "id": "45d53a1e-26e8-4c43-a010-80d141acc249",
                "key": "reasoning",
                "type": "STRING",
            },
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "f0eea82b-39cc-44e3-9c0d-12205ed5652c",
        "type": "ENTRYPOINT",
        "base": None,
        "definition": None,
        "inputs": [],
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "13d9eb34-aecb-496d-9e57-d5e786b0bc7c",
        },
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    subworkflow_node = workflow_raw_data["nodes"][1]
    assert subworkflow_node == {
        "id": "bb98a2c4-c9a7-4c39-8f31-dc7961dc9996",
        "type": "SUBWORKFLOW",
        "inputs": [
            {
                "id": "8107cec2-8215-4730-b52c-859e87a1c116",
                "key": "city",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "693cc9a5-8d74-4a58-bdcf-2b4989cdf250"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "3487e51a-e7fe-4b2c-a1f9-f72c83a329db",
                "key": "date",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "19a78824-9a98-4ae8-a1fc-61f81a422a17"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Example Subworkflow Deployment Node",
            "error_output_id": None,
            "source_handle_id": "ff99bf0c-c239-4b8b-8ac1-483b134f94f4",
            "target_handle_id": "d6194ccf-d31b-4846-8e24-3e189d84351a",
            "variant": "DEPLOYMENT",
            "workflow_deployment_id": deployment_id,
            "release_tag": "LATEST",
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "module": ["vellum", "workflows", "nodes", "displayable", "subworkflow_deployment_node", "node"],
            "name": "SubworkflowDeploymentNode",
        },
        "definition": {
            "module": ["tests", "workflows", "basic_subworkflow_deployment", "workflow"],
            "name": "ExampleSubworkflowDeploymentNode",
        },
        "trigger": {
            "id": "d6194ccf-d31b-4846-8e24-3e189d84351a",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "ff99bf0c-c239-4b8b-8ac1-483b134f94f4", "name": "default", "type": "DEFAULT"}],
        "outputs": [
            {
                "id": "d901cbed-9905-488c-be62-e2668f85438f",
                "name": "temperature",
                "schema": {"type": "number"},
                "type": "NUMBER",
                "value": None,
            },
            {
                "id": "68de689c-fe8a-4189-b7d0-82c620ac30f9",
                "name": "reasoning",
                "schema": {"type": "string"},
                "type": "STRING",
                "value": None,
            },
        ],
    }

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicSubworkflowDeploymentWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_subworkflow_deployment",
            "workflow",
        ],
    }


def test_serialize_workflow__subworkflow_deployment_node_outputs_from_release(vellum_client):
    """
    Tests that subworkflow deployment node outputs are populated from the deployment's output variables,
    and that downstream nodes referencing those outputs correctly reference the subworkflow node.
    """

    # GIVEN a subworkflow deployment node with foo and bar outputs
    class SubworkflowDeploymentNodeWithOutputs(SubworkflowDeploymentNode):
        deployment = "test_subworkflow_deployment"

        class Outputs(BaseOutputs):
            foo: str
            bar: int

    # AND a templating node that references both outputs
    class ConsumingNode(TemplatingNode):
        template = "foo: {{ foo }}, bar: {{ bar }}"
        inputs = {
            "foo": SubworkflowDeploymentNodeWithOutputs.Outputs.foo,
            "bar": SubworkflowDeploymentNodeWithOutputs.Outputs.bar,
        }

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = SubworkflowDeploymentNodeWithOutputs >> ConsumingNode

        class Outputs(BaseOutputs):
            result = ConsumingNode.Outputs.result

    # AND a deployment release with output variables for foo and bar
    foo_output_id = "11111111-1111-1111-1111-111111111111"
    bar_output_id = "22222222-2222-2222-2222-222222222222"
    deployment_release = WorkflowDeploymentRelease(
        id=str(uuid4()),
        created=datetime.now(),
        environment=ReleaseEnvironment(id=str(uuid4()), name="DEVELOPMENT", label="Development"),
        workflow_version=WorkflowDeploymentReleaseWorkflowVersion(
            id=str(uuid4()),
            input_variables=[],
            output_variables=[
                VellumVariable(id=foo_output_id, key="foo", type="STRING"),
                VellumVariable(id=bar_output_id, key="bar", type="NUMBER"),
            ],
        ),
        deployment=WorkflowDeploymentReleaseWorkflowDeployment(
            id=str(uuid4()),
            name="test_subworkflow_deployment",
        ),
        release_tags=[],
        reviews=[],
    )
    vellum_client.workflow_deployments.retrieve_workflow_deployment_release.return_value = deployment_release

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the subworkflow deployment node should have outputs populated from the deployment's output variables
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    subworkflow_node = next(node for node in workflow_raw_data["nodes"] if node["type"] == "SUBWORKFLOW")

    # AND the outputs should contain foo and bar with the correct IDs from the deployment
    assert subworkflow_node.get("outputs") is not None, "outputs field should not be None"
    outputs = subworkflow_node["outputs"]
    assert len(outputs) == 2

    foo_output = next((o for o in outputs if o["name"] == "foo"), None)
    bar_output = next((o for o in outputs if o["name"] == "bar"), None)

    assert foo_output is not None, "foo output should exist"
    assert bar_output is not None, "bar output should exist"
    assert foo_output["id"] == foo_output_id, "foo output ID should match deployment's output variable ID"
    assert bar_output["id"] == bar_output_id, "bar output ID should match deployment's output variable ID"
    assert foo_output["type"] == "STRING"
    assert bar_output["type"] == "NUMBER"

    # AND the consuming node should reference the subworkflow node's outputs
    consuming_node = next(node for node in workflow_raw_data["nodes"] if node["type"] == "TEMPLATING")
    consuming_node_inputs = consuming_node["inputs"]

    foo_input = next((inp for inp in consuming_node_inputs if inp["key"] == "foo"), None)
    bar_input = next((inp for inp in consuming_node_inputs if inp["key"] == "bar"), None)

    assert foo_input is not None, "foo input should exist on consuming node"
    assert bar_input is not None, "bar input should exist on consuming node"

    # AND the foo input should reference the subworkflow node
    foo_rule = foo_input["value"]["rules"][0]
    assert foo_rule["type"] == "NODE_OUTPUT"
    assert foo_rule["data"]["node_id"] == subworkflow_node["id"]
    # The output_id should match the foo output from the subworkflow node's outputs list
    assert foo_rule["data"]["output_id"] == foo_output["id"]

    # AND the bar input should reference the subworkflow node
    bar_rule = bar_input["value"]["rules"][0]
    assert bar_rule["type"] == "NODE_OUTPUT"
    assert bar_rule["data"]["node_id"] == subworkflow_node["id"]
    # The output_id should match the bar output from the subworkflow node's outputs list
    assert bar_rule["data"]["output_id"] == bar_output["id"]


def test_serialize_workflow__wrapped_subworkflow_deployment_node_calls_build_once(vellum_client):
    """
    Tests that a SubworkflowDeploymentNode wrapped in TryNode only calls the deployment
    release API once during serialization, and that the serialized output contains the
    correct deployment_id and release_tag.
    """

    # GIVEN a SubworkflowDeploymentNode wrapped in TryNode
    @TryNode.wrap()
    class WrappedSubworkflowDeploymentNode(SubworkflowDeploymentNode):
        deployment = "test_wrapped_subworkflow_deployment"
        release_tag = "LATEST"

        class Outputs(BaseOutputs):
            result: str

    # AND a workflow that uses the wrapped node
    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = WrappedSubworkflowDeploymentNode

        class Outputs(BaseOutputs):
            final_result = WrappedSubworkflowDeploymentNode.Outputs.result

    # AND a deployment release with specific deployment_id and release_tag
    deployment_id = "test-deployment-id-12345"
    result_output_id = "result-output-id-67890"
    deployment_release = WorkflowDeploymentRelease(
        id=str(uuid4()),
        created=datetime.now(),
        environment=ReleaseEnvironment(id=str(uuid4()), name="DEVELOPMENT", label="Development"),
        workflow_version=WorkflowDeploymentReleaseWorkflowVersion(
            id=str(uuid4()),
            input_variables=[],
            output_variables=[
                VellumVariable(id=result_output_id, key="result", type="STRING"),
            ],
        ),
        deployment=WorkflowDeploymentReleaseWorkflowDeployment(
            id=deployment_id,
            name="test_wrapped_subworkflow_deployment",
        ),
        release_tags=[],
        reviews=[],
    )
    vellum_client.workflow_deployments.retrieve_workflow_deployment_release.return_value = deployment_release

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the deployment release API should only be called once
    assert vellum_client.workflow_deployments.retrieve_workflow_deployment_release.call_count == 1

    # AND the serialized subworkflow node should have the correct deployment_id
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    subworkflow_node = next(node for node in workflow_raw_data["nodes"] if node["type"] == "SUBWORKFLOW")
    assert subworkflow_node["data"]["workflow_deployment_id"] == deployment_id

    # AND the serialized subworkflow node should have the correct release_tag
    assert subworkflow_node["data"]["release_tag"] == "LATEST"



import warnings
from typing import Any, ClassVar, Dict, Generic, Optional, TypeVar, Union, cast

from vellum.workflows.nodes import NoteNode
from vellum.workflows.types.core import JsonObject
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.utils import raise_if_descriptor
from vellum_ee.workflows.display.types import WorkflowDisplayContext

_NoteNodeType = TypeVar("_NoteNodeType", bound=NoteNode)


class BaseNoteNodeDisplay(BaseNodeDisplay[_NoteNodeType], Generic[_NoteNodeType]):
    # Deprecated: Define text and style on the node class instead. Will be removed in v2.0.0.
    text: ClassVar[str] = ""
    style: ClassVar[Union[Dict[str, Any], None]] = None

    def serialize(self, display_context: WorkflowDisplayContext, **kwargs: Any) -> JsonObject:
        del kwargs  # Unused parameters
        node_id = self.node_id
        node = self._node

        text = raise_if_descriptor(node.text) or ""
        style = cast(Optional[Dict[str, Any]], raise_if_descriptor(node.style)) or None

        if "text" in self.__class__.__dict__ or "style" in self.__class__.__dict__:
            warnings.warn(
                "Defining 'text' and 'style' on the display class is deprecated. "
                "Define them on the node class instead. Will be removed in v2.0.0.",
                DeprecationWarning,
                stacklevel=2,
            )

        return {
            "id": str(node_id),
            "type": "NOTE",
            "inputs": [],
            "data": {
                "label": self.label,
                "text": text,
                "style": style,
            },
            **self.serialize_generic_fields(display_context, exclude=["outputs"]),
        }



from unittest.mock import ANY

from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_inline_subworkflow.workflow import BasicInlineSubworkflowWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicInlineSubworkflowWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "fa73da37-34c3-47a9-be58-69cc6cdbfca5",
                "key": "city",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
                "schema": {"type": "string"},
            },
            {
                "id": "aba1e6e0-dfa7-4c15-a4e6-aec6feebfaca",
                "key": "date",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
                "schema": {"type": "string"},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "99afb757-2782-465d-ab55-80ccf50552b9",
                "key": "temperature",
                "type": "NUMBER",
            },
            {
                "id": "7444a019-081a-4e10-a528-3249299159f7",
                "key": "reasoning",
                "type": "STRING",
            },
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "6358dcfe-b162-4e19-99ca-401d1ada9bdc",
        "type": "ENTRYPOINT",
        "base": None,
        "definition": None,
        "inputs": [],
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "c344fdee-282b-40c9-8c97-6dd08830948c",
        },
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    subworkflow_node = workflow_raw_data["nodes"][1]
    assert not DeepDiff(
        {
            "id": "cc246984-82bb-4da3-b447-ea66aa1cffe4",
            "type": "SUBWORKFLOW",
            "inputs": [
                {
                    "id": "9ae1bf55-2024-48e4-b7fa-29c189545dc8",
                    "key": "metro",
                    "value": {
                        "rules": [
                            {
                                "type": "INPUT_VARIABLE",
                                "data": {"input_variable_id": "fa73da37-34c3-47a9-be58-69cc6cdbfca5"},
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "data": {
                "label": "Example Inline Subworkflow Node",
                "error_output_id": None,
                "source_handle_id": "4e6146c9-05f1-4494-b48c-7ff02224cf5a",
                "target_handle_id": "955e83a5-6a19-4dfa-ab19-9a6fd63d017d",
                "variant": "INLINE",
                "workflow_raw_data": ANY,
                "input_variables": [
                    {
                        "id": "9ae1bf55-2024-48e4-b7fa-29c189545dc8",
                        "key": "metro",
                        "type": "STRING",
                        "required": True,
                        "default": None,
                        "schema": {"type": "string"},
                    }
                ],
                "output_variables": [
                    {"id": "2fc57139-7420-49e5-96a6-dcbb3ff5d622", "key": "temperature", "type": "NUMBER"},
                    {"id": "fad5dd9f-3328-4e70-ad55-65a5325a4a82", "key": "reasoning", "type": "STRING"},
                ],
            },
            "display_data": {"position": {"x": 200.0, "y": -50.0}},
            "base": {
                "name": "InlineSubworkflowNode",
                "module": ["vellum", "workflows", "nodes", "core", "inline_subworkflow_node", "node"],
            },
            "definition": {
                "name": "ExampleInlineSubworkflowNode",
                "module": ["tests", "workflows", "basic_inline_subworkflow", "workflow"],
            },
            "trigger": {
                "id": "955e83a5-6a19-4dfa-ab19-9a6fd63d017d",
                "merge_behavior": "AWAIT_ATTRIBUTES",
            },
            "ports": [{"id": "4e6146c9-05f1-4494-b48c-7ff02224cf5a", "name": "default", "type": "DEFAULT"}],
        },
        subworkflow_node,
        exclude_regex_paths=r"root\['data'\]\['workflow_raw_data'\]",
    )

    nested_workflow_data = subworkflow_node["data"]["workflow_raw_data"]
    nested_start_node = next(
        node for node in nested_workflow_data["nodes"] if node["type"] == "GENERIC" and node["label"] == "Start Node"
    )
    assert nested_start_node == {
        "id": "45ba1e29-611f-4a6f-9f51-7cf6fe47141e",
        "label": "Start Node",
        "type": "GENERIC",
        "should_file_merge": True,
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
        "definition": {
            "name": "StartNode",
            "module": ["tests", "workflows", "basic_inline_subworkflow", "workflow"],
        },
        "trigger": {
            "id": "ab273dfa-6719-427c-9054-dd2c03d90d5d",
            "merge_behavior": "AWAIT_ATTRIBUTES",
        },
        "ports": [{"id": "37a41308-2f6c-443d-a832-5859cef231e8", "name": "default", "type": "DEFAULT"}],
        "adornments": None,
        "attributes": [
            {
                "id": "47970f44-eafc-4526-bd81-b1419b0a787a",
                "name": "metro",
                "value": {
                    "type": "WORKFLOW_INPUT",
                    "input_variable_id": "f2f5da15-026d-4905-bfe7-7d16bda20eed",
                },
            },
            {
                "id": "6d933d5e-84a2-434a-baf8-25315c9469af",
                "name": "date",
                "value": {
                    "type": "WORKFLOW_INPUT",
                    "input_variable_id": "aba1e6e0-dfa7-4c15-a4e6-aec6feebfaca",
                },
            },
        ],
        "outputs": [
            {
                "id": "1162f98c-fc51-49c3-a634-120ccbf2ddc2",
                "name": "temperature",
                "type": "NUMBER",
                "value": None,
                "schema": {"type": "number"},
            },
            {
                "id": "4d13651c-3c74-4bf4-8595-2339de83c59a",
                "name": "reasoning",
                "type": "STRING",
                "value": None,
                "schema": {"type": "string"},
            },
        ],
    }

    assert nested_workflow_data["display_data"] == {"viewport": {"x": 0.0, "y": 0.0, "zoom": 1.0}}
    assert nested_workflow_data["definition"] == {
        "name": "NestedWorkflow",
        "module": ["tests", "workflows", "basic_inline_subworkflow", "workflow"],
    }
    assert nested_workflow_data["output_values"] == [
        {
            "output_variable_id": "2fc57139-7420-49e5-96a6-dcbb3ff5d622",
            "value": {
                "type": "NODE_OUTPUT",
                "node_id": "45ba1e29-611f-4a6f-9f51-7cf6fe47141e",
                "node_output_id": "1162f98c-fc51-49c3-a634-120ccbf2ddc2",
            },
        },
        {
            "output_variable_id": "fad5dd9f-3328-4e70-ad55-65a5325a4a82",
            "value": {
                "type": "NODE_OUTPUT",
                "node_id": "45ba1e29-611f-4a6f-9f51-7cf6fe47141e",
                "node_output_id": "4d13651c-3c74-4bf4-8595-2339de83c59a",
            },
        },
    ]

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicInlineSubworkflowWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_inline_subworkflow",
            "workflow",
        ],
    }



import pytest

from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes.displayable.inline_prompt_node.node import InlinePromptNode
from vellum.workflows.nodes.displayable.tool_calling_node.node import ToolCallingNode
from vellum_ee.workflows.display.utils.exceptions import UnsupportedSerializationException
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_serialize_workflow__code_tool_with_node_reference__raises_error():
    """
    Tests that a serialization error is raised when a code tool references a workflow node.
    """

    # GIVEN a function that references a node class
    def my_tool_with_node_reference(query: str) -> str:
        return str(InlinePromptNode)

    # AND a tool calling node that uses this function
    class MyToolCallingNode(ToolCallingNode):
        functions = [my_tool_with_node_reference]

    # AND a workflow with the tool calling node
    class TestWorkflow(BaseWorkflow):
        graph = MyToolCallingNode

        class Outputs(BaseWorkflow.Outputs):
            result = MyToolCallingNode.Outputs.text

    # WHEN we try to serialize the workflow
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)

    # THEN a serialization error should be raised
    with pytest.raises(UnsupportedSerializationException) as exc_info:
        workflow_display.serialize()

    # AND the error message should mention the node class and provide guidance
    error_message = str(exc_info.value)
    assert "InlinePromptNode" in error_message
    assert "Code tools cannot reference workflow nodes" in error_message
    assert "Inline Subworkflow tool" in error_message


def test_serialize_workflow__code_tool_without_node_reference__no_error():
    """
    Tests that no error is raised when a code tool does not reference any workflow nodes,
    even if the module has node imports at the top level.
    """

    # GIVEN a function that does NOT reference any node class (even though InlinePromptNode
    # is imported at the module level)
    def my_simple_tool(query: str) -> str:
        return query.upper()

    # AND a tool calling node that uses this function
    class MyToolCallingNode(ToolCallingNode):
        functions = [my_simple_tool]

    # AND a workflow with the tool calling node
    class TestWorkflow(BaseWorkflow):
        graph = MyToolCallingNode

        class Outputs(BaseWorkflow.Outputs):
            result = MyToolCallingNode.Outputs.text

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)
    serialized_workflow = workflow_display.serialize()

    # THEN no error should be raised and the workflow should serialize successfully
    assert serialized_workflow is not None
    assert "workflow_raw_data" in serialized_workflow


def test_serialize_workflow__code_tool_with_node_instantiation__raises_error():
    """
    Tests that a serialization error is raised when a code tool instantiates a node and calls its run method.
    """

    # GIVEN a function that instantiates a node and calls its run method
    def my_tool():
        node: InlinePromptNode = InlinePromptNode()
        return node.run()

    # AND a tool calling node that uses this function
    class MyToolCallingNode(ToolCallingNode):
        functions = [my_tool]

    # AND a workflow with the tool calling node
    class TestWorkflow(BaseWorkflow):
        graph = MyToolCallingNode

        class Outputs(BaseWorkflow.Outputs):
            result = MyToolCallingNode.Outputs.text

    # WHEN we try to serialize the workflow
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)

    # THEN a serialization error should be raised
    with pytest.raises(UnsupportedSerializationException) as exc_info:
        workflow_display.serialize()

    # AND the error message should mention the node class
    error_message = str(exc_info.value)
    assert "InlinePromptNode" in error_message
    assert "Code tools cannot reference workflow nodes" in error_message



import pytest
from datetime import datetime
from typing import List, Optional, Union

from vellum import (
    ArrayChatMessageContentItem,
    ChatMessage,
    SearchResult,
    VellumAudio,
    VellumDocument,
    VellumImage,
    VellumValue,
)
import vellum.client.types as vellum_types
from vellum.workflows.types.core import Json
from vellum.workflows.utils.vellum_variables import (
    _is_vellum_value_subtype,
    primitive_type_to_vellum_variable_type,
    vellum_variable_type_to_openapi_type,
)


@pytest.mark.parametrize(
    "type_, expected",
    [
        (str, "STRING"),
        (Optional[str], "STRING"),
        (datetime, "STRING"),
        (Optional[datetime], "STRING"),
        (int, "NUMBER"),
        (Optional[int], "NUMBER"),
        (float, "NUMBER"),
        (Optional[float], "NUMBER"),
        (List[ChatMessage], "CHAT_HISTORY"),
        (Optional[List[ChatMessage]], "CHAT_HISTORY"),
        (List[SearchResult], "SEARCH_RESULTS"),
        (Optional[List[SearchResult]], "SEARCH_RESULTS"),
        (Json, "JSON"),
        (Optional[Json], "JSON"),
        (VellumDocument, "DOCUMENT"),
        (Optional[VellumDocument], "DOCUMENT"),
        (VellumAudio, "AUDIO"),
        (Optional[VellumAudio], "AUDIO"),
        (VellumImage, "IMAGE"),
        (Optional[VellumImage], "IMAGE"),
        (list[ChatMessage], "CHAT_HISTORY"),
        (Optional[list[ChatMessage]], "CHAT_HISTORY"),
        (list[SearchResult], "SEARCH_RESULTS"),
        (Optional[list[SearchResult]], "SEARCH_RESULTS"),
        (list[VellumValue], "ARRAY"),
        (Optional[list[VellumValue]], "ARRAY"),
        (List[ArrayChatMessageContentItem], "ARRAY"),
        (Optional[List[ArrayChatMessageContentItem]], "ARRAY"),
    ],
)
def test_primitive_type_to_vellum_variable_type(type_, expected):
    assert primitive_type_to_vellum_variable_type(type_) == expected


@pytest.mark.parametrize(
    "vellum_type, expected",
    [
        ("STRING", "string"),
        ("NUMBER", "number"),
        ("JSON", "object"),
        ("CHAT_HISTORY", "array"),
        ("SEARCH_RESULTS", "array"),
        ("ERROR", "object"),
        ("ARRAY", "array"),
        ("FUNCTION_CALL", "object"),
        ("IMAGE", "object"),
        ("AUDIO", "object"),
        ("DOCUMENT", "object"),
        ("NULL", "null"),
    ],
)
def test_vellum_variable_type_to_openapi_type(vellum_type, expected):
    assert vellum_variable_type_to_openapi_type(vellum_type) == expected


class _SomeType:
    pass


_STRING_OR_NUMBER_VELLUM_VALUE = Union[vellum_types.StringVellumValue, vellum_types.NumberVellumValue]


@pytest.mark.parametrize(
    ["type_", "expected_is_vellum_value_subtype"],
    [
        # Individual union members
        (vellum_types.StringVellumValue, True),
        (vellum_types.NumberVellumValue, True),
        (vellum_types.JsonVellumValue, True),
        (vellum_types.AudioVellumValue, True),
        (vellum_types.VideoVellumValue, True),
        (vellum_types.ImageVellumValue, True),
        (vellum_types.DocumentVellumValue, True),
        (vellum_types.FunctionCallVellumValue, True),
        (vellum_types.ErrorVellumValue, True),
        (vellum_types.ArrayVellumValue, True),
        (vellum_types.ChatHistoryVellumValue, True),
        (vellum_types.SearchResultsVellumValue, True),
        (vellum_types.ThinkingVellumValue, True),
        # Misc union types
        (Union[vellum_types.StringVellumValue, vellum_types.NumberVellumValue], True),
        (_STRING_OR_NUMBER_VELLUM_VALUE, True),
        (Union[vellum_types.StringVellumValue], True),
        (Union["vellum_types.StringVellumValue"], False),
        # Random types
        (int, False),
        (_SomeType, False),
        (List[int], False),
        (Union[int, str], False),
        (Optional[int], False),
        (List["_SomeType"], False),
        # Handle non-types gracefully
        (1, False),
        (_SomeType(), False),
    ],
)
def test_is_vellum_value_subtype(type_, expected_is_vellum_value_subtype):
    assert _is_vellum_value_subtype(type_) == expected_is_vellum_value_subtype



from datetime import datetime
from uuid import uuid4

from deepdiff import DeepDiff

from vellum import WorkspaceSecretRead
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_api_node.workflow import SimpleAPIWorkflow

# 0d76e1e1-3a4b-4eb4-a606-f73d62c -> a0542dcc-443c-4b3b-aac8-c41d2277a5c7


def test_serialize_workflow(vellum_client):
    # GIVEN a Workflow that uses a vellum API node
    # AND stubbed out API calls
    workspace_secret_id = str(uuid4())
    workspace_secret = WorkspaceSecretRead(
        id=workspace_secret_id,
        modified=datetime.now(),
        name="MY_SECRET",
        label="My Secret",
        secret_type="USER_DEFINED",
    )
    vellum_client.workspace_secrets.retrieve.return_value = workspace_secret

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=SimpleAPIWorkflow)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 0

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 3
    assert not DeepDiff(
        [
            {"id": "9a37bf7d-484e-4725-903e-f3254df38a0a", "key": "json", "type": "JSON"},
            {"id": "5090e96d-5787-4a08-bf58-129101cf2548", "key": "headers", "type": "JSON"},
            {"id": "44ea8d75-e2a8-4627-85b1-8504b65d25c9", "key": "status_code", "type": "NUMBER"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND the API node should be serialized correctly
    api_node = next(n for n in workflow_raw_data["nodes"] if (n.get("base") or {}).get("name") == "APINode")

    assert api_node["id"] == "6783c80f-5fc0-4712-a251-ce59d3c14ff2"
    assert api_node["type"] == "API"

    assert api_node["base"]["name"] == "APINode"
    assert api_node["base"]["module"] == ["vellum", "workflows", "nodes", "displayable", "api_node", "node"]

    assert api_node["definition"]["name"] == "SimpleAPINode"
    assert api_node["definition"]["module"] == ["tests", "workflows", "basic_api_node", "workflow"]

    assert api_node["trigger"]["id"] == "35816b8f-453b-4f70-a5fc-72dd0ceca460"
    assert api_node["trigger"]["merge_behavior"] == "AWAIT_ANY"

    assert len(api_node["ports"]) == 1
    assert api_node["ports"][0]["name"] == "default"
    assert api_node["ports"][0]["type"] == "DEFAULT"

    assert len(api_node["outputs"]) == 4
    output_names = {output["name"] for output in api_node["outputs"]}
    assert output_names == {"json", "headers", "status_code", "text"}

    assert api_node["data"]["label"] == "Simple API Node"
    assert api_node["data"]["error_output_id"] is None
    assert api_node["data"]["source_handle_id"] == "6c574f01-9362-4edd-b3dd-5faacca76b28"
    assert api_node["data"]["target_handle_id"] == "35816b8f-453b-4f70-a5fc-72dd0ceca460"
    assert api_node["data"]["json_output_id"] == "a0542dcc-443c-4b3b-aac8-c41d2277a5c7"
    assert api_node["data"]["text_output_id"] == "f7190a6b-0c13-4a5a-9087-d8e6feb84eca"
    assert api_node["data"]["status_code_output_id"] == "f687c0a1-b63a-4c5c-b6ef-472c6108ae4b"

    input_keys = {inp["key"] for inp in api_node["inputs"]}
    assert "url" in input_keys
    assert "method" in input_keys
    assert "body" in input_keys

    assert len(api_node["attributes"]) == 1
    assert api_node["attributes"][0]["name"] == "timeout"

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "SimpleAPIWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_api_node",
            "workflow",
        ],
    }



"""Pydantic models for workflow metadata.json files.

The metadata.json file is generated alongside workflow code to ensure stable,
UI-aligned identifiers without relying on Python display classes. It contains
mappings between code paths and UI identifiers for nodes, triggers, edges, and
dataset rows.
"""

from typing import Dict, Optional

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.client.types.code_resource_definition import CodeResourceDefinition


class RunnerConfig(UniversalBaseModel):
    """Configuration for the workflow runner environment.

    This specifies the container image and version information used to execute
    the workflow.
    """

    container_image_name: Optional[str] = None
    """The name of the container image to use. If not provided, defaults to the
    standard Python workflow runtime image."""

    container_image_tag: Optional[str] = None
    """The tag of the container image to use. If not provided, defaults to the
    latest tag."""

    codegen_version: Optional[str] = None
    """The version of codegen used to generate the workflow artifact.

    .. deprecated::
        This field is deprecated. Use `sdk_version` instead.
    """

    sdk_version: Optional[str] = None
    """The SDK version the workflow is intended to run on."""

    is_deployment_inlining_enabled: bool = False
    """Whether deployment inlining is enabled for this workflow."""

    server_version: Optional[str] = None
    """The server version the workflow is intended to run on. This gets set when
    deploying a workflow with a hotswappable custom image."""


class WorkflowMetadata(UniversalBaseModel):
    """Metadata persisted alongside generated workflow code.

    This metadata ensures stable, UI-aligned identifiers without relying on
    Python display classes. It is generated by the codegen system and consumed
    during workflow serialization to maintain consistent IDs across round-trips.
    """

    runner_config: Optional[RunnerConfig] = None
    """Configuration for the workflow runner environment."""

    label: Optional[str] = None
    """Human-readable label for the workflow."""

    codegen_version: Optional[str] = None
    """The version of codegen used to generate this workflow.

    .. deprecated::
        This field is deprecated. Use `runner_config.codegen_version` instead.
    """

    server_version: Optional[str] = None
    """The server version this workflow was generated for.

    .. deprecated::
        This field is deprecated. Use `runner_config.server_version` instead.
    """

    node_id_to_file_mapping: Dict[str, CodeResourceDefinition] = {}
    """Mapping of node IDs to their code resource definitions.

    Key: Node UUID string
    Value: CodeResourceDefinition containing the module path and class name
    """

    trigger_path_to_id_mapping: Dict[str, str] = {}
    """Mapping of trigger class paths to their UI trigger IDs.

    Key: ".<relative_trigger_module_path>.<TriggerClassName>"
    Value: UI trigger ID (UUID string)
    """

    trigger_attribute_id_mapping: Dict[str, str] = {}
    """Mapping of trigger attribute keys to their UI attribute IDs.

    Key: "<trigger_path>|<attribute_key>"
    Value: Attribute ID (UUID string)

    This ensures trigger attribute IDs remain stable across serialization
    round-trips.
    """

    entrypoint: Optional[str] = None
    """The UI entrypoint node ID.

    Kept for compatibility. Edge IDs for entrypoint flows are now included in
    edges_to_id_mapping using the manual trigger path as the source.
    """

    edges_to_id_mapping: Dict[str, str] = {}
    """Mapping of edge path keys to their UI edge IDs.

    Key: "<source_path>|<target_node_path>"
        where:
            source_path := "<trigger_module_path>.<TriggerClassName>" for trigger edges
                          or "<module_path>.<NodeClassName>.Ports.<source_handle_id>"
                             for node->node edges (handle id disambiguates multiple edges)
            target_node_path := "<module_path>.<ClassName>.Trigger" for the target node
    Value: UI edge ID (UUID string)
    """

    dataset_row_index_to_id_mapping: Dict[str, str] = {}
    """Mapping of dataset row indices to their UI row IDs.

    Key: Dataset row index (0-based) as a string
    Value: UI dataset row ID (UUID string)
    """



from typing import Any

from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import BaseNode
from vellum.workflows.nodes.displayable.final_output_node import FinalOutputNode
from vellum.workflows.state.base import BaseState
from vellum_ee.workflows.display.utils.exceptions import UnsupportedSerializationException
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_final_output_node_display__serialize_with_valid_types():
    # GIVEN a node that outputs a str
    class StringNode(BaseNode):

        class Outputs:
            result: str

    # AND a FinalOutputNode with matching type to that node
    class CorrectOutput(FinalOutputNode[BaseState, str]):
        class Outputs(FinalOutputNode.Outputs):
            value = StringNode.Outputs.result

    # AND a workflow referencing the node
    class MyWorkflow(BaseWorkflow):
        graph = StringNode >> CorrectOutput

        class Outputs(BaseWorkflow.Outputs):
            final_result = CorrectOutput.Outputs.value

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)

    # THEN serialization should succeed without raising validation errors
    serialized_workflow: dict = workflow_display.serialize()

    # AND the node should be properly serialized
    assert "workflow_raw_data" in serialized_workflow
    assert "nodes" in serialized_workflow["workflow_raw_data"]

    # Find the terminal node in the serialized output
    terminal_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] == "TERMINAL"
    )
    assert terminal_node is not None
    assert terminal_node["id"] == str(CorrectOutput.__id__)


def test_final_output_node_display__serialize_with_invalid_types_should_raise_error():
    # GIVEN a node that outputs a str
    class StringNode(BaseNode):

        class Outputs:
            result: str

    # AND a FinalOutputNode with mismatched types (expects list but gets str)
    class BadOutput(FinalOutputNode[BaseState, list]):
        """Output with type mismatch."""

        class Outputs(FinalOutputNode.Outputs):
            value = StringNode.Outputs.result  # str type, conflicts with list

    # AND a workflow referencing the node
    class MyWorkflow(BaseWorkflow):
        graph = StringNode >> BadOutput

        class Outputs(BaseWorkflow.Outputs):
            final_result = BadOutput.Outputs.value

    # WHEN we attempt to serialize the workflow
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)

    # THEN serialization should complete without raising an exception
    serialized_workflow = workflow_display.serialize()

    # AND the error should be captured in workflow_display.errors
    errors = list(workflow_display.display_context.errors)
    assert len(errors) == 1
    assert "Output type mismatch" in str(errors[0])
    assert "list" in str(errors[0])
    assert "str" in str(errors[0])

    # AND the serialized workflow should still be created
    assert "workflow_raw_data" in serialized_workflow


def test_final_output_node_display__serialize_with_list_str_type():
    """
    Tests that FinalOutputNode with list[str] as second type parameter serializes correctly.
    """

    # GIVEN a node that outputs a list of strings
    class ListNode(BaseNode):

        class Outputs:
            result: list[str]

    # AND a FinalOutputNode with list[str] as the second type parameter
    class ListOutput(FinalOutputNode[BaseState, list[str]]):
        class Outputs(FinalOutputNode.Outputs):
            value = ListNode.Outputs.result

    # AND a workflow referencing the node
    class MyWorkflow(BaseWorkflow):
        graph = ListNode >> ListOutput

        class Outputs(BaseWorkflow.Outputs):
            final_result = ListOutput.Outputs.value

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN serialization should succeed without raising validation errors
    assert "workflow_raw_data" in serialized_workflow
    assert "nodes" in serialized_workflow["workflow_raw_data"]

    # AND the terminal node should be properly serialized
    terminal_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] == "TERMINAL"
    )
    assert terminal_node is not None
    assert terminal_node["id"] == str(ListOutput.__id__)

    # AND the output type should be correctly serialized as JSON
    assert terminal_node["data"]["output_type"] == "JSON"

    # AND the outputs should contain the correct type information
    assert len(terminal_node["outputs"]) == 1
    assert terminal_node["outputs"][0]["type"] == "JSON"

    # AND the output should have an inline type reference with JSON Schema
    schema = terminal_node["outputs"][0]["schema"]
    assert schema["type"] == "array"
    assert schema["items"]["type"] == "string"


def test_final_output_node_display__serialize_with_nested_node_output_reference():
    """
    Tests that FinalOutputNode with a dict containing nested node output references serializes correctly.
    """

    # GIVEN a node that outputs multiple values
    class DataNode(BaseNode):
        class Outputs:
            url: str
            count: int

    # AND a FinalOutputNode with a dict containing nested references to node outputs
    class NestedOutput(FinalOutputNode[BaseState, dict[str, Any]]):
        class Outputs(FinalOutputNode.Outputs):
            value = {
                "download_url": DataNode.Outputs.url,
                "row_count": DataNode.Outputs.count,
                "message": "Data processed successfully",
            }

    # AND a workflow referencing the node
    class MyWorkflow(BaseWorkflow):
        graph = DataNode >> NestedOutput

        class Outputs(BaseWorkflow.Outputs):
            final_result = NestedOutput.Outputs.value

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN serialization should succeed
    assert "workflow_raw_data" in serialized_workflow
    assert "nodes" in serialized_workflow["workflow_raw_data"]

    # AND the terminal node should be properly serialized
    terminal_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] == "TERMINAL"
    )
    assert terminal_node is not None
    assert terminal_node["id"] == str(NestedOutput.__id__)

    # AND the node should have inputs
    assert len(terminal_node["inputs"]) == 1
    node_input = terminal_node["inputs"][0]
    assert node_input["key"] == "node_input"

    # AND the input value should have rules with a CONSTANT_VALUE containing the serialized dict
    assert node_input["value"]["combinator"] == "OR"
    assert len(node_input["value"]["rules"]) == 1
    rule = node_input["value"]["rules"][0]
    assert rule["type"] == "CONSTANT_VALUE"
    assert rule["data"]["type"] == "JSON"

    # AND the serialized value should be a DICTIONARY_REFERENCE with entries
    serialized_value = rule["data"]["value"]
    assert serialized_value["type"] == "DICTIONARY_REFERENCE"
    assert "entries" in serialized_value
    assert len(serialized_value["entries"]) == 3

    # AND the entries should contain the nested references and constant value
    entry_keys = {entry["key"] for entry in serialized_value["entries"]}
    assert entry_keys == {"download_url", "row_count", "message"}

    # AND the outputs should contain the DICTIONARY_REFERENCE value
    assert len(terminal_node["outputs"]) == 1
    output = terminal_node["outputs"][0]
    assert output["name"] == "value"
    assert output["type"] == "JSON"
    assert output["value"]["type"] == "DICTIONARY_REFERENCE"
    assert len(output["value"]["entries"]) == 3

    # AND the attributes should be empty for this node
    assert terminal_node.get("attributes", []) == []

    # AND a validation error should be recorded for the nested references
    errors = list(workflow_display.display_context.errors)
    assert len(errors) == 1
    assert isinstance(errors[0], UnsupportedSerializationException)
    assert "nested references" in str(errors[0]).lower()



"""Tests for IntegrationTrigger with explicit entrypoint_node_id serialization."""

from uuid import uuid4

from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.state.base import BaseState
from vellum.workflows.triggers.integration import IntegrationTrigger
from vellum_ee.workflows.display.base import EdgeDisplay, EntrypointDisplay, WorkflowMetaDisplay
from vellum_ee.workflows.display.editor.types import NodeDisplayData
from vellum_ee.workflows.display.workflows.base_workflow_display import BaseWorkflowDisplay
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_integration_trigger_with_explicit_entrypoint_node_id():
    """
    Tests that a workflow with an IntegrationTrigger and explicit entrypoint_node_id
    does NOT create an ENTRYPOINT node when all branches are trigger-sourced.

    Even with explicit IDs provided, if all graph branches originate from triggers,
    the ENTRYPOINT node should be skipped.
    """

    # GIVEN an IntegrationTrigger workflow with explicit entrypoint_node_id
    class SlackMessageTrigger(IntegrationTrigger):
        message: str

        class Config:
            provider = "COMPOSIO"
            integration_name = "SLACK"
            slug = "slack_message"

    class ProcessNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result = SlackMessageTrigger.message

        def run(self) -> Outputs:
            return self.Outputs()

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = SlackMessageTrigger >> ProcessNode

    explicit_entrypoint_node_id = uuid4()
    explicit_entrypoint_source_handle_id = uuid4()

    class TestWorkflowDisplay(BaseWorkflowDisplay[TestWorkflow]):
        workflow_display = WorkflowMetaDisplay(
            entrypoint_node_id=explicit_entrypoint_node_id,
            entrypoint_node_source_handle_id=explicit_entrypoint_source_handle_id,
            entrypoint_node_display=NodeDisplayData(),
        )
        entrypoint_displays = {
            ProcessNode: EntrypointDisplay(
                id=uuid4(),
                edge_display=EdgeDisplay(id=uuid4()),
            )
        }

    # WHEN we serialize the workflow
    result: dict = get_workflow_display(workflow_class=TestWorkflow).serialize()

    # THEN the trigger should be serialized
    assert "workflow_raw_data" in result
    workflow_raw_data = result["workflow_raw_data"]
    assert isinstance(workflow_raw_data, dict)

    assert "triggers" in result
    triggers = result["triggers"]
    assert isinstance(triggers, list)
    assert len(triggers) == 1
    trigger = triggers[0]
    assert isinstance(trigger, dict)
    trigger_id = trigger["id"]

    # AND there should be NO ENTRYPOINT node (all branches are trigger-sourced)
    nodes = workflow_raw_data["nodes"]
    assert isinstance(nodes, list)

    entrypoint_nodes = [n for n in nodes if isinstance(n, dict) and n.get("type") == "ENTRYPOINT"]

    assert len(entrypoint_nodes) == 0, (
        "IntegrationTrigger-only workflows should NOT create an ENTRYPOINT node "
        "even with explicit entrypoint_node_id, since all branches are trigger-sourced"
    )

    # AND edges should use trigger ID as source_node_id
    edges = workflow_raw_data["edges"]
    assert isinstance(edges, list)
    trigger_edges = [e for e in edges if isinstance(e, dict) and e.get("source_node_id") == trigger_id]
    assert len(trigger_edges) > 0, "Should have edges from trigger ID"



import pytest

from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import BaseNode
from vellum.workflows.references.lazy import LazyReference


class ResponseNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        response: str = "Hello from node!"


class TestWorkflowWithOutput(BaseWorkflow):
    graph = ResponseNode

    class Outputs(BaseWorkflow.Outputs):
        final_response = ResponseNode.Outputs.response


class TestWorkflowWithLiteralOutput(BaseWorkflow):
    graph = ResponseNode

    class Outputs(BaseWorkflow.Outputs):
        literal_value = "Hello literal!"


@pytest.fixture
def mock_inspect_getsource(mocker):
    return mocker.patch("inspect.getsource")


@pytest.fixture
def mock_logger(mocker):
    return mocker.patch("vellum.workflows.references.lazy.logger")


def test_lazy_reference__inspect_getsource_fails(mock_inspect_getsource, mock_logger):
    # GIVEN getsource fails to resolve the lambda's source code
    mock_inspect_getsource.side_effect = Exception("test")

    # WHEN a node with a lazy reference is defined
    class MyNode(BaseNode):
        lazy_reference = LazyReference(lambda: MyNode.Outputs.foo)

    # THEN the name is the lambda function's name
    assert MyNode.lazy_reference.instance
    assert MyNode.lazy_reference.instance.name == "<lambda>"

    # AND sentry is notified
    assert mock_logger.exception.call_count == 1


def test_lazy_reference__string_resolves_workflow_output():
    """Tests that string-based LazyReference can resolve workflow output references."""

    # GIVEN a workflow with an output that references a node output (defined at module level)
    # WHEN we run the workflow
    terminal_event = TestWorkflowWithOutput().run()

    # THEN the workflow completes successfully
    assert terminal_event.name == "workflow.execution.fulfilled"
    assert terminal_event.final_state is not None

    # AND we can resolve a string-based LazyReference to the workflow output
    lazy_ref = LazyReference[str]("TestWorkflowWithOutput.Outputs.final_response")
    resolved_value = lazy_ref.resolve(terminal_event.final_state)

    # THEN the resolved value matches the node output
    assert resolved_value == "Hello from node!"


def test_lazy_reference__string_resolves_literal_workflow_output():
    """Tests that string-based LazyReference can resolve literal workflow output values."""

    # GIVEN a workflow with a literal output value (defined at module level)
    # WHEN we run the workflow
    terminal_event = TestWorkflowWithLiteralOutput().run()

    # THEN the workflow completes successfully
    assert terminal_event.name == "workflow.execution.fulfilled"
    assert terminal_event.final_state is not None

    # AND we can resolve a string-based LazyReference to the literal workflow output
    lazy_ref = LazyReference[str]("TestWorkflowWithLiteralOutput.Outputs.literal_value")
    resolved_value = lazy_ref.resolve(terminal_event.final_state)

    # THEN the resolved value matches the literal output
    assert resolved_value == "Hello literal!"



"""Tests for IntegrationTrigger attribute type validation during serialization."""

import pytest
from unittest.mock import MagicMock, patch
from typing import Dict

from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.state.base import BaseState
from vellum.workflows.triggers.integration import IntegrationTrigger
from vellum_ee.workflows.display.utils.exceptions import TriggerValidationError
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


class LinearIssueTrigger(IntegrationTrigger):
    """Custom Linear issue trigger for testing."""

    action: str
    data: dict
    type: str
    url: str

    class Config:
        provider = "COMPOSIO"
        integration_name = "LINEAR"
        slug = "LINEAR_ISSUE_CREATED_TRIGGER"
        setup_attributes = {"team_id": "test-team-id"}


class ProcessNode(BaseNode):
    """Node that processes the trigger."""

    class Outputs(BaseNode.Outputs):
        result = LinearIssueTrigger.action

    def run(self) -> Outputs:
        return self.Outputs()


class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
    graph = LinearIssueTrigger >> ProcessNode


def _create_mock_tool_definition(properties: Dict[str, dict]) -> MagicMock:
    """Helper to create a mock tool definition with output_parameters as JSON Schema.

    For triggers, output_parameters contains the webhook payload schema,
    while input_parameters contains setup/config arguments.
    """
    mock_tool_def = MagicMock()
    mock_tool_def.name = "LINEAR_ISSUE_CREATED_TRIGGER"
    # output_parameters is a JSON Schema object containing the payload schema
    mock_tool_def.output_parameters = {
        "type": "object",
        "properties": properties,
        "required": list(properties.keys()),
    }
    return mock_tool_def


def test_integration_trigger_validation__matching_types():
    """
    Tests that serialization succeeds when trigger attribute types match the API definition.
    """
    # GIVEN a tool definition from the API with matching types (JSON Schema format)
    mock_tool_def = _create_mock_tool_definition(
        {
            "action": {"type": "string"},
            "data": {"type": "object"},
            "type": {"type": "string"},
            "url": {"type": "string"},
        }
    )

    # WHEN we serialize the workflow with mocked API
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)

    with patch.object(
        workflow_display._client.integrations,
        "retrieve_integration_tool_definition",
        return_value=mock_tool_def,
    ):
        # THEN serialization should succeed without raising an error
        result = workflow_display.serialize()

    # AND the trigger should be serialized correctly
    assert "triggers" in result
    triggers = result["triggers"]
    assert isinstance(triggers, list)
    assert len(triggers) == 1


def test_integration_trigger_validation__type_mismatch():
    """
    Tests that serialization raises TriggerValidationError when attribute types don't match.
    """
    # GIVEN a tool definition from the API with mismatched types (JSON Schema format)
    # The API says 'action' should be object (JSON), but our trigger defines it as STRING
    mock_tool_def = _create_mock_tool_definition(
        {
            "action": {"type": "object"},
            "data": {"type": "object"},
            "type": {"type": "string"},
            "url": {"type": "string"},
        }
    )

    # WHEN we serialize the workflow with mocked API
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)

    with patch.object(
        workflow_display._client.integrations,
        "retrieve_integration_tool_definition",
        return_value=mock_tool_def,
    ):
        # THEN serialization should raise TriggerValidationError
        with pytest.raises(TriggerValidationError) as exc_info:
            workflow_display.serialize()

    # AND the error message should indicate the type mismatch with expected and actual types
    assert "action" in str(exc_info.value)
    assert "STRING" in str(exc_info.value)
    assert "expected type 'JSON'" in str(exc_info.value)
    assert "The trigger configuration is invalid or contains unsupported values" in str(exc_info.value)


def test_integration_trigger_validation__api_error():
    """
    Tests that serialization continues gracefully when the API returns an error.
    """
    # GIVEN an API that raises an exception
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)

    with patch.object(
        workflow_display._client.integrations,
        "retrieve_integration_tool_definition",
        side_effect=Exception("API Error"),
    ):
        # WHEN we serialize the workflow
        # THEN serialization should succeed (validation is skipped on API error)
        result = workflow_display.serialize()

    # AND the trigger should still be serialized
    assert "triggers" in result
    triggers = result["triggers"]
    assert isinstance(triggers, list)
    assert len(triggers) == 1


def test_integration_trigger_validation__no_properties():
    """
    Tests that serialization continues gracefully when the API response has no properties.
    """
    # GIVEN a tool definition with empty properties in JSON Schema
    mock_tool_def = _create_mock_tool_definition({})

    # WHEN we serialize the workflow with mocked API
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)

    with patch.object(
        workflow_display._client.integrations,
        "retrieve_integration_tool_definition",
        return_value=mock_tool_def,
    ):
        # THEN serialization should succeed (validation is skipped when no output_parameters)
        result = workflow_display.serialize()

    # AND the trigger should still be serialized
    assert "triggers" in result
    triggers = result["triggers"]
    assert isinstance(triggers, list)
    assert len(triggers) == 1



from typing import TYPE_CHECKING, Any, List, Optional, Union

from vellum.client.types import (
    ArrayChatMessageContent,
    ArrayChatMessageContentItem,
    AudioChatMessageContent,
    ChatMessage,
    ChatMessageContent,
    DocumentChatMessageContent,
    FunctionCallChatMessageContent,
    ImageChatMessageContent,
    StringChatMessageContent,
    VideoChatMessageContent,
)
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.descriptors.utils import resolve_value
from vellum.workflows.references.lazy import LazyReference
from vellum.workflows.references.output import OutputReference
from vellum.workflows.triggers.base import BaseTrigger
from vellum.workflows.utils.pydantic_schema import validate_obj_as

if TYPE_CHECKING:
    from vellum.workflows.state.base import BaseState


class ChatMessageTrigger(BaseTrigger):
    """
    Trigger for chat-based workflows that supports multi-turn conversations.

    Appends user message to state.chat_history at workflow start, and assistant
    response after workflow completion. Use previous_execution_id to maintain
    conversation state across executions.

    Attributes:
        message: The incoming chat message content. Can be a string or a list of content items.
    """

    message: Union[str, List[ArrayChatMessageContentItem]]

    class Config(BaseTrigger.Config):
        output: Optional[BaseDescriptor[Any]] = None

    def __init__(self, **kwargs: Any):
        """Initialize ChatMessageTrigger, converting VellumValue objects to ChatMessageContent if needed."""
        # Convert message from VellumValue format to ChatMessageContent format if needed
        if "message" in kwargs:
            message = kwargs["message"]
            # Handle string messages by converting to a list with a single StringChatMessageContent
            if isinstance(message, str):
                kwargs["message"] = [StringChatMessageContent(value=message)]
            elif isinstance(message, list):
                converted_message = []
                for item in message:
                    # If it's already a ChatMessageContent type, keep it as-is
                    if isinstance(
                        item,
                        (
                            StringChatMessageContent,
                            ImageChatMessageContent,
                            AudioChatMessageContent,
                            VideoChatMessageContent,
                            DocumentChatMessageContent,
                            FunctionCallChatMessageContent,
                        ),
                    ):
                        converted_message.append(item)
                    # Handle raw strings in the array by wrapping them in StringChatMessageContent
                    elif isinstance(item, str):
                        converted_message.append(StringChatMessageContent(value=item))
                    # Convert VellumValue objects or dicts to ChatMessageContent
                    # Use discriminated union validation
                    else:
                        # Get the dict representation (either from Pydantic model or already a dict)
                        item_dict = item.model_dump() if hasattr(item, "model_dump") else item
                        converted_message.append(
                            validate_obj_as(ArrayChatMessageContentItem, item_dict)  # type: ignore[arg-type]
                        )

                kwargs["message"] = converted_message

        super().__init__(**kwargs)

    def __on_workflow_initiated__(self, state: "BaseState") -> None:
        """Appends user message to state.chat_history at workflow start."""
        if not hasattr(state, "chat_history"):
            return

        if state.chat_history is None:
            state.chat_history = []

        if isinstance(self.message, str):
            user_message = ChatMessage(
                role="USER",
                text=self.message,
            )
        else:
            user_message = ChatMessage(
                role="USER",
                content=ArrayChatMessageContent(value=self.message),
            )
        state.chat_history.append(user_message)

    def __on_workflow_fulfilled__(self, state: "BaseState") -> None:
        """Appends assistant response to state.chat_history after workflow completion."""
        if not hasattr(state, "chat_history"):
            return

        if state.chat_history is None:
            state.chat_history = []

        output = self.Config.output
        if output is not None:
            resolved_output = self._resolve_output(output, state)
            assistant_message = ChatMessage(
                role="ASSISTANT",
                content=resolved_output if not isinstance(resolved_output, str) else None,
                text=resolved_output if isinstance(resolved_output, str) else None,
            )
            state.chat_history.append(assistant_message)

    def _resolve_output(
        self,
        output: BaseDescriptor[Any],
        state: "BaseState",
    ) -> Union[str, ChatMessageContent]:
        """Resolves output reference, handling workflow output references."""
        descriptor = output
        if isinstance(output, LazyReference) and callable(output._get):
            descriptor = output._get()

        if isinstance(descriptor, OutputReference) and isinstance(descriptor.instance, BaseDescriptor):
            return resolve_value(descriptor.instance, state)

        return resolve_value(output, state)

    class Display(BaseTrigger.Display):
        label: str = "Chat Message"
        icon: Optional[str] = "vellum:icon:message-dots"
        color: Optional[str] = "blue"



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_templating_node.workflow_with_json_input import BasicTemplatingNodeWorkflowWithJson


def test_serialize_workflow():
    # GIVEN a Workflow that uses a vellum templating node",

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicTemplatingNodeWorkflowWithJson)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert input_variables == [
        {
            "id": "f4435f88-703f-40e5-9197-d39b0e43ab72",
            "key": "info",
            "type": "JSON",
            "default": None,
            "required": True,
            "extensions": {"color": None},
            "schema": {},
        }
    ]

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert not DeepDiff(
        [{"id": "62ec9b08-6437-4f8d-bc20-983d317bc348", "key": "result", "type": "JSON"}],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND the templating node should be serialized correctly
    templating_node = next(
        n for n in workflow_raw_data["nodes"] if (n.get("base") or {}).get("name") == "TemplatingNode"
    )

    assert templating_node["id"] == "28ac8611-a755-4d8a-a5c3-520ddd119cf8"
    assert templating_node["type"] == "TEMPLATING"

    assert templating_node["base"]["name"] == "TemplatingNode"
    assert templating_node["base"]["module"] == ["vellum", "workflows", "nodes", "core", "templating_node", "node"]

    assert templating_node["definition"]["name"] == "ExampleTemplatingNode"
    assert templating_node["definition"]["module"] == [
        "tests",
        "workflows",
        "basic_templating_node",
        "workflow_with_json_input",
    ]

    assert templating_node["trigger"]["id"] == "a215a440-cbd9-48ae-a831-be8fa78530a6"
    assert templating_node["trigger"]["merge_behavior"] == "AWAIT_ATTRIBUTES"

    assert len(templating_node["ports"]) == 1
    assert templating_node["ports"][0]["name"] == "default"
    assert templating_node["ports"][0]["type"] == "DEFAULT"

    assert templating_node["data"]["label"] == "Example Templating Node"
    assert templating_node["data"]["output_id"] == "076becd5-f282-40c5-9917-61099e114298"
    assert templating_node["data"]["error_output_id"] is None
    assert templating_node["data"]["source_handle_id"] == "90a2ea33-b08d-46ec-8d9e-612592764268"
    assert templating_node["data"]["target_handle_id"] == "a215a440-cbd9-48ae-a831-be8fa78530a6"
    assert templating_node["data"]["template_node_input_id"] == "7ceb3528-49c6-4f90-b3ef-92b1921a1b7d"
    assert templating_node["data"]["output_type"] == "JSON"

    input_keys = {inp["key"] for inp in templating_node["inputs"]}
    assert "template" in input_keys
    assert "info" in input_keys

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicTemplatingNodeWorkflowWithJson",
        "module": ["tests", "workflows", "basic_templating_node", "workflow_with_json_input"],
    }
